# -*- coding: UTF-8 -*-
"""
    grailkit.util
    ~~~~~~~~~~~~~

    Utility functions and constants

    :copyright: (c) 2017-2019 by Oleksii Lytvyn.
    :license: MIT, see LICENSE for more details.
"""
import os
import sys
import time
import types
import shutil
import platform

# Platform constants
OS_WIN = platform.system() == "Windows"
OS_MAC = platform.system() == "Darwin"
OS_UNIX = platform.system() == "unix"
OS_LINUX = platform.system() == "Linux"

BUILTIN_TYPES = (
    bool,
    int,
    float,
    complex,
    str,
    dict,
    list,
    set,
    frozenset,
    tuple,
    bytes,
    bytearray,
    types.FunctionType,
    types.LambdaType,
    types.BuiltinFunctionType
    )


def application_location():
    """Get the path to main script file no matter how it's run.

    Returns:
        str: path to current python script
    """

    if hasattr(sys, 'frozen'):
        path = os.path.dirname(sys.executable)
    elif '__file__' in locals():
        path = os.path.dirname(__file__)
    else:
        path = sys.path[0]

    return os.path.abspath(path)


def data_location(app_name):
    """Returns path to application data folder

    Args:
        app_name (str): name of application

    Returns:
        str: path to folder
    """

    if sys.platform == 'win32':
        path = os.path.join(os.environ['APPDATA'], app_name)
    else:
        path = os.path.expanduser(os.path.join("~", "." + app_name))

    return os.path.abspath(path)


def copy_file(source, destination):
    """Copy file 'source' to file 'destination'

    Args:
        source (str): path of file to be copied
        destination (str): file destination path

    Returns:
        bool: True if copy is successful otherwise False
    """

    directory = os.path.dirname(os.path.realpath(destination))

    if not os.path.exists(source):
        return False

    if not os.path.exists(directory):
        os.makedirs(directory)

    if os.path.exists(source):
        shutil.copyfile(source, destination)

    return True


def millis_now():
    """Get current time in ms

    Returns:
        int: time in milliseconds since epoch
    """

    return int(round(time.time() * 1000))


def default_key(obj, key, default=None):
    """Get value by attribute in object or by key in dict
    if property exists returns value otherwise return `default` value.
    If object not exist or object don't have property returns `default`

    Args:
        obj (object, list, tuple, dict): dictionary
        key (str, int): key of property
        default (object): Object that returned if key not found
    Returns:
        value by given key or default
    """

    if obj and isinstance(obj, dict):
        return obj[key] if key in obj else default
    elif obj and isinstance(obj, object):
        return getattr(obj, key, default)
    else:
        return default


def file_exists(path):
    """Check if file exists

    Args:
        path (str): path to file
    Returns:
        bool: True if file exists, False otherwise
    """

    return os.path.exists(os.path.dirname(os.path.realpath(path))) and os.path.isfile(path)


def is_builtin(type_object):
    """Check if object type is built-in

    Args:
        type_object (type, object): object or type to check
    Returns:
        bool: True if built-in else False
    """

    return isinstance(type_object, type) or isinstance(type_object, BUILTIN_TYPES)


def object_type(object_ref):
    """Get type of anything

    Args:
        object_ref (object, type): Object or type
    Returns:
        type: Type of given object
    """

    return object_ref if isinstance(object_ref, type) else type(object_ref)




# -*- coding: UTF-8 -*-
"""
    grailkit.osc
    ~~~~~~~~~~~~

    OSC protocol implementation in pure python

    :copyright: (c) 2017-2019 by Oleksii Lytvyn.
    :license: MIT, see LICENSE for more details.
"""

import re
import time
import struct
import socket
import logging
import decimal
import datetime
import builtins
import calendar
import socketserver

__version__ = '0.6.3'
__all__ = [
    'OSCType',
    'OSCPacket',
    'OSCMessage',
    'OSCBundle',
    'OSCClient',
    'OSCServer',

    'OSCImpulse',
    'OSCColor',
    'OSCMidi',
    'IMMEDIATELY',

    'NTPError',
    'OSCParseError',
    'OSCBuildError'
    ]


class NTPError(Exception):
    """Base class for ntp errors."""


class OSCParseError(Exception):
    """Exception raised when a datagram parsing error occurs."""


class OSCBuildError(Exception):
    """Exception raised  when a datagram building error occurs."""


NTP_IMMEDIATELY = struct.pack('>q', 1)
_NTP_SYSTEM_EPOCH = datetime.date(*time.gmtime(0)[0:3])
_NTP_EPOCH = datetime.date(1900, 1, 1)
_NTP_DELTA = (_NTP_SYSTEM_EPOCH - _NTP_EPOCH).days * 24 * 3600


def ntp_to_time(date):
    """Convert a NTP time to system time.
    System time is represented by seconds since the epoch in UTC.

    Args:
        date: NTP time to be converted
    Returns:
        system time in seconds
    """

    return date - _NTP_DELTA


def time_to_ntp(date):
    """Convert a system time to a NTP time datagram.
    System time is represented by seconds since the epoch in UTC.

    Args:
        date: System time to be converted
    Returns:
        NTP time in seconds
    Raises:
        NTPError if date is invalid
    """
    try:
        ntp = date + _NTP_DELTA
    except TypeError as ve:
        raise NTPError('Invalid date: {}'.format(ve))

    num_secs, fraction = str(ntp).split('.')

    return struct.pack('>I', int(num_secs)) + struct.pack('>I', int(fraction))


IMMEDIATELY = 0

_INT_DGRAM_LEN = 4
_UINT_DGRAM_LEN = 4
_BLOB_DGRAM_PAD = 4
_FLOAT_DGRAM_LEN = 4
_STRING_DGRAM_PAD = 4
_TIMETAG_DGRAM_LEN = _INT_DGRAM_LEN * 2
_COLOR_DGRAM_LEN = 4
_MIDI_DGRAM_LEN = 4
_DOUBLE_DGRAM_LEN = 8
_INT64_DGRAM_LEN = 8
_CHAR_DGRAM_LEN = 4


class OSCImpulse(object):
    """Representation of Impulse OSC type"""

    pass


class OSCMidi(object):
    """Representation of OSC OSCMidi message"""

    __slots__ = ['port', 'status', 'data1', 'data2']

    def __init__(self, port, status, data1=None, data2=None):
        self.port = port
        self.status = status
        self.data1 = data1
        self.data2 = data2

    def pack(self):
        """Returns midi message representation"""
        return struct.pack('>BBBB', self.port, self.status, self.data1, self.data2)

    @classmethod
    def unpack(cls, data):
        """Parse datagram into OSCMidi instance

        Args:
            data (bytes): datagram

        Returns:
            OSCMidi instance
        """
        return cls(*struct.unpack('>BBBB', data))


class OSCColor(object):
    """Representation of OSC color type in RGBA format"""

    __slots__ = ['r', 'g', 'b', 'a']

    def __init__(self, r, g, b, a):
        self.r = r
        self.g = g
        self.b = b
        self.a = a

    def pack(self):
        """Returns datagram of OSCColor type"""

        return struct.pack('>BBBB', self.r, self.g, self.b, self.a)

    @classmethod
    def unpack(cls, data):
        """Parse datagram into OSCColor instance

        Args:
            data (bytes): datagram

        Returns:
            OSCColor instance
        """

        return cls(*struct.unpack('>BBBB', data))


class OSCType(object):
    """Reading and writing OSC types"""

    TYPE_INT = 'i'
    TYPE_UINT = 'u'
    TYPE_CHAR = 'c'
    TYPE_BLOB = 'b'
    TYPE_MIDI = 'm'
    TYPE_NULL = 'N'
    TYPE_TRUE = 'T'
    TYPE_FALSE = 'F'
    TYPE_FLOAT = 'f'
    TYPE_COLOR = 'r'
    TYPE_STRING = 's'
    TYPE_IMPULSE = 'I'
    TYPE_TIMETAG = 't'
    TYPE_DOUBLE = 'd'
    TYPE_INT64 = 'h'
    TYPE_UTF8_STRING = 'S'

    # list of all supported types
    _SUPPORTED_TYPES = (
        TYPE_INT,
        TYPE_UINT,
        TYPE_CHAR,
        TYPE_BLOB,
        TYPE_MIDI,
        TYPE_NULL,
        TYPE_TRUE,
        TYPE_FALSE,
        TYPE_FLOAT,
        TYPE_COLOR,
        TYPE_STRING,
        TYPE_IMPULSE,
        TYPE_TIMETAG,

        TYPE_INT64,
        TYPE_DOUBLE,
        TYPE_UTF8_STRING
        )

    # map of types and corresponding io methods
    TYPES_MAP = {
        TYPE_INT: 'int',
        TYPE_UINT: 'uint',
        TYPE_CHAR: 'char',
        TYPE_BLOB: 'blob',
        TYPE_MIDI: 'midi',
        TYPE_FLOAT: 'float',
        TYPE_COLOR: 'color',
        TYPE_STRING: 'string',
        TYPE_TIMETAG: 'timetag',
        TYPE_DOUBLE: 'double',
        TYPE_INT64: 'int64',
        TYPE_UTF8_STRING: 'utf8_string'
        }

    @classmethod
    def is_supported(cls, _type):
        """Check if given type is supported

        Args:
            _type (str): OSC type tag
        """

        return _type in cls._SUPPORTED_TYPES

    @classmethod
    def tag(cls, value):
        """Get OSC type tag for `value` argument

        Args:
            value: argument value

        Returns:
            OSC type tag
        """

        builtin_type = type(value)
        arg_type = cls.TYPE_NULL

        if builtin_type == builtins.str:
            arg_type = cls.TYPE_STRING
        elif builtin_type == builtins.bytes:
            arg_type = cls.TYPE_BLOB
        elif builtin_type == builtins.int:
            arg_type = cls.TYPE_INT
        elif builtin_type == builtins.float:
            arg_type = cls.TYPE_FLOAT
        elif builtin_type == builtins.bool and value:
            arg_type = cls.TYPE_TRUE
        elif builtin_type == builtins.bool and not value:
            arg_type = cls.TYPE_FALSE
        elif value == None:
            arg_type = cls.TYPE_NULL
        elif builtin_type == builtins.str and len(value) == 1:
            arg_type = cls.TYPE_CHAR
        elif isinstance(value, OSCColor):
            arg_type = cls.TYPE_COLOR
        elif isinstance(value, OSCMidi):
            arg_type = cls.TYPE_MIDI
        elif isinstance(value, OSCImpulse):
            arg_type = cls.TYPE_IMPULSE

        return arg_type

    @classmethod
    def has_datagram(cls, _type):
        """Check if this type has i/o method

        Args:
            _type (str): OSC type tag
        Returns
            True if type has datagram processing method
        """

        return _type in cls.TYPES_MAP

    @classmethod
    def type(cls, _type, data, index=-1):
        """Read and write any supported OSC type

        Args:
            _type (str): OSC type tag
            data (bytes, object): datagram or value to be converted
            index (int): index from which type datagram starts in datagram given
        Returns:
            Datagram from value given if `index` is -1, otherwise parse datagram into value
        """

        return getattr(cls, cls.TYPES_MAP[_type])(data, index)

    @classmethod
    def string(cls, data, index=-1):
        """Get a python string from the datagram and vice versa

        According to the specifications, a string is:
        "A sequence of non-null ASCII characters followed by a null,
        followed by 0-3 additional null characters to make the total number
        of bits a multiple of 32".

        Args:
            data: A datagram packet or value to be converted
            index: An index where the string starts in the datagram.
        Returns:
            A tuple containing the string and the new end index.
        Raises:
            OSCParseError if the datagram could not be parsed.
            OSCBuildError if datagram could not be build.
        """

        # read
        if index != -1:
            offset = 0

            try:
                while data[index + offset] != 0:
                    offset += 1

                if offset == 0:
                    raise OSCParseError('OSC string cannot begin with a null byte: %s' % data[index:])

                if offset % _STRING_DGRAM_PAD == 0:
                    offset += _STRING_DGRAM_PAD
                else:
                    offset += (-offset % _STRING_DGRAM_PAD)

                if offset > len(data[index:]):
                    raise OSCParseError('Datagram is too short')

                data_str = data[index:index + offset]

                return data_str.replace(b'\x00', b'').decode('ascii'), index + offset
            except (IndexError, TypeError) as ex:
                raise OSCParseError('Could not parse datagram %s' % ex)
        # write
        else:
            try:
                dgram = data.encode('ascii')  # Default, but better be explicit.
            except (UnicodeEncodeError, AttributeError) as e:
                raise OSCBuildError('Incorrect string, could not encode {}'.format(e))

            diff = _STRING_DGRAM_PAD - (len(dgram) % _STRING_DGRAM_PAD)
            dgram += (b'\x00' * diff)

            return dgram

    @classmethod
    def utf8_string(cls, data, index=-1):
        """Get a python string from the datagram and vice versa
        Like OSC string but in UTF-8 encoding

        Args:
            data: A datagram packet or value to be converted
            index: An index where the string starts in the datagram.
        Returns:
            A tuple containing the string and the new end index.
        Raises:
            OSCParseError if the datagram could not be parsed.
            OSCBuildError if datagram could not be build.
        """

        # read
        if index != -1:
            offset = 0

            try:
                while data[index + offset] != 0:
                    offset += 1

                if offset == 0:
                    raise OSCParseError('OSC string cannot begin with a null byte: %s' % data[index:])

                if offset % _STRING_DGRAM_PAD == 0:
                    offset += _STRING_DGRAM_PAD
                else:
                    offset += (-offset % _STRING_DGRAM_PAD)

                if offset > len(data[index:]):
                    raise OSCParseError('Datagram is too short')

                data_str = data[index:index + offset]

                return data_str.replace(b'\x00', b'').decode('utf-8'), index + offset
            except (IndexError, TypeError) as ex:
                raise OSCParseError('Could not parse datagram %s' % ex)
        # write
        else:
            try:
                dgram = data.encode('utf-8')
            except (UnicodeEncodeError, AttributeError) as e:
                raise OSCBuildError('Incorrect string, could not encode {}'.format(e))

            diff = _STRING_DGRAM_PAD - (len(dgram) % _STRING_DGRAM_PAD)
            dgram += (b'\x00' * diff)

            return dgram

    @classmethod
    def int(cls, data, index=-1):
        """Returns the datagram for the given integer parameter value
        Get a 32-bit big-endian two's complement integer from the datagram

        Args:
            data: A datagram packet or value to be converted
            index: An index where the integer starts in the datagram.
        Returns:
            Datagram of value given if `index` is -1 otherwise
            tuple containing the integer and the new end index.
        Raises:
            OSCParseError if the datagram could not be parsed.
            OSCBuildError if the int could not be converted.
        """

        # read
        if index != -1:
            try:
                if len(data[index:]) < _INT_DGRAM_LEN:
                    raise OSCParseError('Datagram is too short')
                return struct.unpack('>i', data[index:index + _INT_DGRAM_LEN])[0], index + _INT_DGRAM_LEN
            except (struct.error, TypeError) as e:
                raise OSCParseError('Could not parse datagram %s' % e)
        # write
        else:
            try:
                return struct.pack('>i', data)
            except struct.error as e:
                raise OSCBuildError('Wrong argument value passed: {}'.format(e))

    @classmethod
    def uint(cls, data, index=-1):
        """Returns the datagram for the given integer parameter value
        Get a 32-bit big-endian unsigned integer from the datagram

        Args:
            data: A datagram packet or value to be converted
            index: An index where the integer starts in the datagram.
        Returns:
            Datagram of value given if `index` is -1 otherwise
            tuple containing the integer and the new end index.
        Raises:
            OSCParseError if the datagram could not be parsed.
            OSCBuildError if the int could not be converted.
        """

        # read
        if index != -1:
            try:
                if len(data[index:]) < _UINT_DGRAM_LEN:
                    raise OSCParseError('Datagram is too short')
                return struct.unpack('>I', data[index:index + _UINT_DGRAM_LEN])[0], index + _UINT_DGRAM_LEN
            except (struct.error, TypeError) as e:
                raise OSCParseError('Could not parse datagram %s' % e)
        # write
        else:
            try:
                return struct.pack('>I', data)
            except struct.error as e:
                raise OSCBuildError('Wrong argument value passed: {}'.format(e))

    @classmethod
    def int64(cls, data, index=-1):
        """Read and write Int64 from OSC message

        Args:
            data: integer or datagram
            index: An index where the integer starts in the datagram.
        Returns:
            Datagram of int64 if `index` is -1
            Integer parsed from datagram otherwise
        Raises:
            OSCBuildError if data can't be written
            OSCParseError if data could not be parsed from datagram
        """

        # read
        if index != -1:
            try:
                return struct.unpack('>q', data[index:index + _INT64_DGRAM_LEN])[0], index + _INT64_DGRAM_LEN
            except (struct.error, TypeError) as e:
                raise OSCParseError('Could not parse datagram %s' % e)
        # write
        else:
            try:
                return struct.pack('>q', data)
            except struct.error as e:
                raise OSCBuildError('Wrong argument value passed: {}'.format(e))

    @classmethod
    def double(cls, data, index=-1):
        """Read and write OSC double datagram

        Args:
            data: double or datagram of it
            index: An index where the double starts in the datagram.
        Returns:
            Double datagram if `index` is -1
            Double value and next index
        Raises:
            OSCBuildError if `data` can't be packed
            OSCParseError if `data` could not be parsed
        """

        # read
        if index != -1:
            try:
                return struct.unpack('>d', data[index:index + _DOUBLE_DGRAM_LEN])[0], index + _DOUBLE_DGRAM_LEN
            except (struct.error, TypeError) as e:
                raise OSCParseError('Could not parse datagram %s' % e)
        # write
        else:
            try:
                return struct.pack('>d', data)
            except struct.error as e:
                raise OSCBuildError('Wrong argument value passed: {}'.format(e))

    @classmethod
    def float(cls, data, index=-1):
        """Get a 32-bit big-endian IEEE 754 floating point number from the datagram.
        Returns the datagram for the given float parameter value if `index` is -1

        Args:
            data: A datagram packet of value to be converted
            index: An index where the float starts in the datagram.
        Returns:
            A tuple containing the float and the new end index if `index` != -1
            Datagram of float value if `index` is -1
        Raises:
            OSCParseError if the datagram could not be parsed.
            OSCBuildError if the float could not be converted.
        """

        # read
        if index != -1:

            try:
                if len(data[index:]) < _FLOAT_DGRAM_LEN:
                    data += b'\x00' * (_FLOAT_DGRAM_LEN - len(data[index:]))

                return struct.unpack('>f', data[index:index + _FLOAT_DGRAM_LEN])[0], index + _FLOAT_DGRAM_LEN
            except (struct.error, TypeError) as e:
                raise OSCParseError('Could not parse datagram %s' % e)
        # write
        else:
            try:
                return struct.pack('>f', data)
            except struct.error as e:
                raise OSCBuildError('Wrong argument value passed: {}'.format(e))

    @classmethod
    def timetag(cls, data, index=-1):
        """Get a 64-bit big-endian fixed-point time tag as a date from the datagram.
        Create time tag datagram if `index` is -1

        According to the specifications, a date is represented as is:
        "the first 32 bits specify the number of seconds since midnight on
        January 1, 1900, and the last 32 bits specify fractional parts of a second
        to a precision of about 200 picoseconds".

        Args:
            data: A datagram packet or time value
            index: An index where the date starts in the datagram.
        Returns:
            A tuple containing the system date and the new end index.
            returns IMMEDIATELY (0) if the corresponding OSC sequence was found.
            Create time tag datagram if `index` is -1
        Raises:
            OSCParseError if the datagram could not be parsed.
            NTPError if time cant be converted
        """
        # read
        if index != -1:
            # Check for the special case first.
            if data[index:index + _TIMETAG_DGRAM_LEN] == NTP_IMMEDIATELY:
                return IMMEDIATELY, index + _TIMETAG_DGRAM_LEN

            if len(data[index:]) < _TIMETAG_DGRAM_LEN:
                raise OSCParseError('Datagram is too short')

            num_secs, index = OSCType.int(data, index)
            fraction, index = OSCType.int(data, index)
            # Get a decimal representation from those two values.
            dec = decimal.Decimal(str(num_secs) + '.' + str(fraction))
            # And convert it to float simply.
            system_time = float(dec)

            return ntp_to_time(system_time), index
        # write
        else:
            if data == IMMEDIATELY:
                return NTP_IMMEDIATELY

            try:
                return time_to_ntp(data)
            except NTPError as error:
                raise OSCBuildError(error)

    @classmethod
    def color(cls, data, index=-1):
        """Read and write OSCColor OSC type

        Args:
            data (bytes, OSCColor): datagram or OSCColor instance
            index: An index where the type starts in the datagram.
        Returns:
            OSCColor instance if index given
            Datagram for OSCColor instance if `index` is -1
        Raises:
            OSCParseError if the datagram could not be parsed.
            OSCBuildError if datagram for OSCMidi message can't be created
        """

        # read
        if index != -1:
            try:
                return OSCColor.unpack(data[index:index + _COLOR_DGRAM_LEN]), index + _COLOR_DGRAM_LEN
            except (struct.error, TypeError) as e:
                raise OSCParseError('Could not parse datagram %s' % e)
        # write
        else:
            try:
                return data.pack()
            except struct.error as e:
                raise OSCBuildError('Wrong argument value passed: {}'.format(e))

    @classmethod
    def midi(cls, data, index=-1):
        """Read and write OSC OSCMidi message

        Args:
            data: OSCMidi instance or datagram to be parsed
            index: An index where the type starts in the datagram.
        Returns:
            OSCMidi message datagram if `index` is -1
            OSCMidi instance otherwise
        Raises:
            OSCParseError if the datagram could not be parsed.
            OSCBuildError if datagram for OSCMidi message can't be created
        """

        # read
        if index != -1:
            try:
                return OSCMidi.unpack(data[index:index + _MIDI_DGRAM_LEN]), index + _MIDI_DGRAM_LEN
            except (struct.error, TypeError) as e:
                raise OSCParseError('Could not parse datagram %s' % e)
        # write
        else:
            try:
                return data.pack()
            except struct.error as e:
                raise OSCBuildError('Wrong argument value passed: {}'.format(e))

    @classmethod
    def char(cls, data, index=-1):
        """Read and write Char OSC type

        Args:
            data: Datagram or char
            index: An index when data starts in datagram
        Returns:
            Char if `index` was given
            Datagram if `index` is -1
        Raises:
            OSCParseError if the datagram could not be parsed.
            OSCBuildError if datagram can't be created
        """

        # read
        if index != -1:
            try:
                return data[index].decode('ascii'), index + _CHAR_DGRAM_LEN
            except (struct.error, TypeError) as e:
                raise OSCParseError('Could not parse datagram %s' % e)
        # write
        else:
            try:
                return struct.pack('>c', data.encode('ascii')) + b'\x00' * 3
            except (struct.error, UnicodeEncodeError) as e:
                raise OSCBuildError('Wrong argument value passed: {}'.format(e))

    @classmethod
    def blob(cls, data, index=-1):
        """Get a blob from the datagram if index given otherwise
        returns the datagram for the given `data` value.

        According to the specifications, a blob is made of
        "an int32 size count, followed by that many 8-bit bytes of arbitrary
        binary data, followed by 0-3 additional zero bytes to make the total
        number of bits a multiple of 32".

        Args:
            data: A datagram packet.
            index: An index where the float starts in the datagram.
        Returns:
            A tuple containing the blob and the new end index if index was given
            Datagram of blob `data` if `index` is -1
        Raises:
            OSCParseError if the datagram could not be parsed.
            OSCBuildError if the value was empty or if its size didn't fit an OSC int.
        """

        # read
        if index != -1:
            size, offset = OSCType.int(data, index)
            # Make the size a multiple of 32 bits.
            total_size = size + (-size % _BLOB_DGRAM_PAD)
            end_index = offset + size

            if end_index - index > len(data[index:]):
                raise OSCParseError('Datagram is too short.')

            return data[offset:offset + size], offset + total_size
        # write
        else:
            val = data

            if not val:
                raise OSCBuildError('Blob value cannot be empty')

            dgram = OSCType.int(len(val)) + val

            while len(dgram) % _BLOB_DGRAM_PAD != 0:
                dgram += b'\x00'

            return dgram


class OSCPacket(object):
    """Unit of transmission of the OSC protocol.

    Any application that sends OSC Packets is an OSC Client.
    Any application that receives OSC Packets is an OSC Server.
    """

    def __init__(self, dgram):
        """Initialize an OSCPacket with the given UDP datagram.

        Args:
            dgram: the raw UDP datagram holding the OSC packet.
        Raises:
            OSCParseError if the datagram could not be parsed.
        """

        self.time = calendar.timegm(time.gmtime())
        self.dgram = dgram
        self.message = None

        try:
            if OSCBundle.is_valid(dgram):
                self.message = OSCBundle.parse(dgram)
            elif OSCMessage.is_valid(dgram):
                self.message = OSCMessage.parse(dgram)
            else:
                # Empty packet, should not happen as per the spec but heh, UDP...
                raise OSCParseError('OSC Packet should at least contain an OSCMessage or an OSCBundle.')
        except OSCParseError as pe:
            raise OSCParseError('Could not parse packet: %s' % pe)

    def __cmp__(self, other):
        """Compare two OSCPacket's

        Args:
            other (OSCPacket): other OSCPacket to compare
        Returns:
            True if two OSCPacket's is the same
        """

        return self.dgram == other.dgram

    @property
    def size(self):
        """Size of datagram"""

        return len(self.dgram)


class OSCMessage(object):
    """Builds arbitrary OSCMessage instances."""

    def __init__(self, address="/", args=None):
        """Initialize a new OSCMessage.

        Args:
            address (str): The osc address to send this message to.
            args (list): list of args to add to message
        """

        self._address = "/"
        self._args = []
        self._dgram = b''

        # OSC address will be checked here
        self.address = address

        if args and len(args) > 0:
            for value in args:
                self.add(value)

    def __iter__(self):
        """Returns an iterator over the arguments of this message in pairs (osc type tag, value)"""

        return iter(self._args)

    def __len__(self):
        """Returns length of arguments"""

        return len(self._args)

    def __getitem__(self, key):
        """Get a OSCMessage argument by index

        Args:
            key (int): index of argument
        Returns:
            argument of OSCMessage
        Raises:
            IndexError if index out of range
        """

        if key >= len(self._args):
            raise IndexError("Index out of range.")

        return self._args[key]

    def __setitem__(self, key, value):
        """Set argument by index, if key is greater than
        length of arguments error will be raised

        Args:
            key (int): index of argument
            value: an argument
        Raises:
            IndexError if there is no argument with this index
        """

        if key >= len(self._args):
            raise IndexError("Index out of range.")
        else:
            arg_type = OSCType.tag(value)
            self._args[key] = (arg_type, value)

    def __delitem__(self, key):
        """Delete argument by index

        Args:
            key (int): index of argument
        Raises:
            IndexError if there is no argument with this index
        """

        if key >= len(self._args):
            raise IndexError("Index out of range.")

        del self._args[key]

    def __contains__(self, value):
        """Returns True if value in arguments list

        Args:
            value: argument value
        """

        return self.index(value) >= 0

    def __cmp__(self, other):
        """Check two OSCMessage's to be equal

        Args:
            other (OSCMessage): other OSCMessage
        Returns:
            True if two messages equals
        """

        return self.build().dgram == other.build().dgram

    @property
    def address(self):
        """Returns the OSC address this message will be sent to."""

        return self._address

    @address.setter
    def address(self, value):
        """Sets the OSC address this message will be sent to.

        Args:
            value (str): OSC message address pattern
        """

        if not value.startswith('/'):
            raise ValueError("Given '%s' OSC address doesn't start with /." % str(value))
        elif not self.is_valid_address(value):
            raise ValueError("Given '%s' OSC address doesn't matches with valid address pattern." % str(value))

        self._address = value

    @property
    def args(self):
        """Returns list of value added as arguments"""

        return [a[1] for a in self._args]

    @property
    def size(self):
        """Returns length of the datagram for this message."""

        return len(self._dgram)

    @property
    def dgram(self):
        """Returns datagram from which this message was built."""

        return self._dgram

    def add(self, value, _type=None):
        """Add a typed argument to this message.

        Args:
            value: The corresponding value for the argument.
            _type: A value in ARG_TYPE_* defined in this class,
                if none then the type will be guessed.
        Raises:
            ValueError: if the type is not supported.
        """

        self.append(value, _type)

    def append(self, value, _type=None):
        """Add a typed argument to this message.

        Args:
            value: The corresponding value for the argument.
            _type: A value in ARG_TYPE_* defined in this class,
                if none then the type will be guessed.
        Raises:
            ValueError: if the type is not supported.
        """

        if _type and not OSCType.is_supported(_type):
            raise ValueError('Given type is not supported')

        if not _type:
            _type = OSCType.tag(value)

        self._args.append((_type, value))

    def extend(self, values):
        """Extend arguments list, all values will be added using auto type

        Args:
            values (list): a list of values
        """

        for value in values:
            self.append(value)

    def insert(self, index, value, _type=None):
        """Insert typed argument at specific index

        Args:
            index (int): index of insertion
            value: The corresponding value for the argument.
            _type: A value in ARG_TYPE_* defined in this class,
                if none then the type will be guessed.
        Raises:
            ValueError: if the type is not supported.
            IndexError: if index is greater than list size
        """

        if index >= len(self._args):
            raise IndexError("Index is out of range.")

        if _type and not OSCType.is_supported(_type):
            raise ValueError('Given type is not supported')

        if not _type:
            _type = OSCType.tag(value)

        self._args.insert(index, (_type, value))

    def remove(self, value):
        """Remove the first item from the arguments list whose value is `value`.

        Args:
            value: argument value
        Raises:
            ValueError: if value is not fund in arguments list
        """

        index = self.index(value)

        if (index > 0) and (index < len(self._args)):
            del self._args[index]
        else:
            raise ValueError("Item not found in arguments list.")

    def index(self, value, start=0, end=False):
        """Find value in arguments list and return first index otherwise return -1.
        The returned index is computed relative to the beginning of the full sequence rather than the start argument.
        """
        found_index = -1

        if not end:
            end = len(self._args)

        for index, arg in enumerate(self._args):

            if index < start:
                continue

            if index >= end:
                break

            if value == arg[1]:
                found_index = index

        return found_index

    def clear(self):
        """Remove all arguments from message"""

        self._args.clear()

    def copy(self):
        """Create copy of OSCMessage

        Returns:
            New OSCMessage instance
        """
        self.build()

        return OSCMessage.parse(self._dgram)

    def build(self):
        """Builds OSCMessage datagram and return current instance

        Returns:
            an OSCMessage instance.
        Raises:
            OSCBuildError: if the message could not be build or if the address was empty.
        """

        if not self._address:
            raise OSCBuildError('OSC addresses cannot be empty')

        dgram = b''

        try:
            # Write the address.
            dgram += OSCType.string(self._address)

            if not self._args:
                self._dgram = dgram

                return self

            # Write the parameters.
            types = "".join([arg[0] for arg in self._args])
            dgram += OSCType.string(',' + types)

            for _type, value in self._args:
                # if type in list use function to create datagram
                if OSCType.has_datagram(_type):
                    dgram += OSCType.type(_type, value)
                # process arg without datagram
                elif OSCType.is_supported(_type):
                    continue
                # otherwise type not supported
                else:
                    raise OSCBuildError('Incorrect parameter type found {}'.format(_type))

            self._dgram = dgram

            return self
        except OSCBuildError as be:
            raise OSCBuildError('Could not build the message: {}'.format(be))

    def _parse(self, dgram):
        """Parse datagram

        Args:
            dgram (bytes): datagram of OSCMessage
        """

        self._dgram = dgram

        try:
            self._address, index = OSCType.string(self._dgram, 0)

            if not self._dgram[index:]:
                # No params is legit, just return now.
                return

            # Get the parameters types.
            typetag, index = OSCType.string(self._dgram, index)

            if typetag.startswith(','):
                typetag = typetag[1:]

            # Parse each parameter given its type.
            for _type in typetag:

                if OSCType.has_datagram(_type):
                    value, index = OSCType.type(_type, self._dgram, index)
                elif _type == OSCType.TYPE_TRUE:
                    value = True
                elif _type == OSCType.TYPE_FALSE:
                    value = False
                elif _type == OSCType.TYPE_NULL:
                    value = None
                elif _type == OSCType.TYPE_IMPULSE:
                    value = OSCImpulse()
                else:
                    logging.warning('Unhandled parameter type: {0}'.format(_type))
                    continue

                self._args.append((_type, value))
        except OSCParseError as pe:
            raise OSCParseError('Found incorrect datagram, ignoring it', pe)

    @staticmethod
    def parse(dgram):
        """Create OSCMessage from datagram

        Args:
            dgram (bytes): from what to build OSCMessage
        Returns:
            OSCMessage parsed from datagram
        """

        message = OSCMessage()
        message._parse(dgram)

        return message

    @staticmethod
    def is_valid(dgram):
        """Check datagram to be valid OSCMessage

        Args:
            dgram (bytes): datagram os of OSCMessage
        Returns:
            whether this datagram starts as an OSC message.
        """

        return dgram.startswith(b'/')

    @staticmethod
    def is_valid_address(address):
        """Check if given value is valid OSC-string address pattern

        Args:
            address (str): OSC address pattern
        Returns:
            True if valid
        """

        return address == '/' or re.compile("^/[a-zA-Z0-9/_\-?*\[\]]+").match(address)


class OSCBundle(object):
    """Builds arbitrary OSCBundle instances."""

    _BUNDLE_PREFIX = b"#bundle\x00"

    def __init__(self, timestamp=IMMEDIATELY, messages=None):
        """Build a new bundle with the associated timestamp.

        Args:
            timestamp (int): system time represented as a floating point number of
                       seconds since the epoch in UTC or IMMEDIATELY.
            messages (list): List of OSCMessage's to add to bundle
        """

        self._timestamp = timestamp
        self._contents = []
        self._dgram = b''

        if messages and len(messages) > 0:
            for value in messages:
                self.add(value)

    def __iter__(self):
        """Returns an iterator over the bundle's content."""

        return iter(self._contents)

    def __len__(self):
        """Returns length of contents"""

        return len(self._contents)

    def __getitem__(self, key):
        """Get item from OSCBundle by index

        Args:
            key (int): index of item
        Returns:
            item from contents
        """

        return self._contents[key]

    def __setitem__(self, key, value):
        """Set item of OSCBundle

        Args:
            key (int): index of item
            value (OSCBundle, OSCMessage): an OSCBundle or OSCMessage
        """

        if key >= len(self._contents):
            raise IndexError("Index out of range.")

        if not isinstance(value, OSCBundle) or not isinstance(value, OSCMessage):
            raise TypeError("Type of assigned values is not OSCBundle or OSCMessage.")

        self._contents[key] = value

    def __delitem__(self, key):
        """Remove item from bundle

        Args:
            key (int): index of item
        """

        if key >= len(self._contents):
            raise IndexError("Index out of range.")

        del self._contents[key]

    def __contains__(self, item):
        """Check if OSCMessage in bundle

        Returns:
            returns True if item in this bundle
        """

        return item in self._contents

    def __cmp__(self, other):
        """Compare two bundles

        Args:
            other (OSCBundle): other bundle to compare
        """

        return self.dgram == other.dgram

    @property
    def timestamp(self):
        """Returns timestamp associated with this bundle."""

        return self._timestamp

    @property
    def length(self):
        """Returns number of messages in bundle."""

        return len(self._contents)

    @property
    def size(self):
        """Returns length of the datagram for this bundle."""

        return len(self._dgram)

    @property
    def dgram(self):
        """Returns datagram from which this bundle was built."""

        return self._dgram

    def append(self, content):
        """Add a new content to this bundle.

        Args:
            content: Either an OSCBundle or an OSCMessage
        Raises:
            OSCBuildError: if we could not build the bundle.
        """

        if isinstance(content, OSCMessage) or isinstance(content, OSCBundle):
            self._contents.append(content.build())
        else:
            raise OSCBuildError("Content must be either OSCBundle or OSCMessage found {}".format(type(content)))

    def add(self, content):
        """Same as append method"""

        self.append(content)

    def build(self):
        """Build an OSCBundle with the current state of this builder.

        Raises:
            OSCBuildError: if we could not build the bundle.
        """

        dgram = b'' + self._BUNDLE_PREFIX

        try:
            dgram += OSCType.timetag(self._timestamp)

            for content in self._contents:
                if isinstance(content, OSCMessage) or isinstance(content, OSCBundle):
                    size = content.size
                    dgram += OSCType.int(size)
                    dgram += content.dgram
                else:
                    raise OSCBuildError(
                        "Content must be either OSCBundle or OSCMessage found {}".format(type(content)))

            return OSCBundle.parse(dgram)
        except OSCBuildError as be:
            raise OSCBuildError('Could not build the bundle {}'.format(be))

    @classmethod
    def is_valid(cls, dgram):
        """Returns whether this datagram starts like an OSC bundle.

        Args:
            dgram: datagram of OSCBundle
        Returns:
            weather datagram is OSCBundle
        """

        return dgram.startswith(cls._BUNDLE_PREFIX)

    @staticmethod
    def parse(dgram):
        """Parse OSCBundle from datagram

        Args:
            dgram: datagram of OSCBundle
        Returns:
            OSCBundle instance
        """

        bundle = OSCBundle()
        bundle._parse(dgram)

        return bundle

    def _parse(self, dgram):
        """Parse datagram and fill contents of this OSCBundle

        Args:
            dgram (bytes): datagram of OSCBundle
        """

        # Interesting stuff starts after the initial b"#bundle\x00".
        self._dgram = dgram
        index = len(self._BUNDLE_PREFIX)

        try:
            self._timestamp, index = OSCType.timetag(self._dgram, index)
        except OSCParseError as pe:
            raise OSCParseError("Could not get the date from the datagram: %s" % pe)

        # Get the contents as a list of OSCBundle and OSCMessage.
        self._contents = self._parse_contents(index)

    def _parse_contents(self, index):
        """Parse datagram into OSCBundle

        Args:
            index (int): start index of next OSCMessage in bundle
        Raises:
            OSCParseError: if we could not parse the bundle.
        """

        contents = []

        try:
            # An OSC Bundle Element consists of its size and its contents.
            # The size is an int32 representing the number of 8-bit bytes in the
            # contents, and will always be a multiple of 4. The contents are either
            # an OSC Message or an OSC Bundle.
            while self._dgram[index:]:
                # Get the sub content size.
                content_size, index = OSCType.int(self._dgram, index)
                # Get the datagram for the sub content.
                content_dgram = self._dgram[index:index + content_size]
                # Increment our position index up to the next possible content.
                index += content_size
                # Parse the content into an OSC message or bundle.
                if OSCBundle.is_valid(content_dgram):
                    contents.append(OSCBundle.parse(content_dgram))
                elif OSCMessage.is_valid(content_dgram):
                    contents.append(OSCMessage.parse(content_dgram))
                else:
                    logging.warning("Could not identify content type of dgram %s" % content_dgram)
        except (OSCParseError, IndexError) as e:
            raise OSCParseError("Could not parse a content datagram: %s" % e)

        return contents


class OSCClient(object):
    """Send OSCMessage's and OSCBundle's to multiple servers"""

    def __init__(self, address='127.0.0.1', port=False):
        """Initialize the client.

        Args:
            address (str): recipient ip address
            port (int): recipient port
        """

        self._socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self._socket.setblocking(0)
        self._clients = []
        self._closed = False

        if address and port:
            self.add(address, port)

    def __len__(self):
        """Returns number of clients"""

        return len(self._clients)

    def __bool__(self):
        """Returns True if at least one client is available"""

        return len(self) > 0

    @property
    def clients(self):
        """Returns list of receipts"""

        return self._clients

    def add(self, address, port):
        """Add a recipient

        Args:
            address (str): ip address of server
            port (int): port of server
        Raises:
            ValueError if one of arguments is invalid
        """

        if not isinstance(address, str):
            raise ValueError("Given address is not a string")

        if not isinstance(port, int) or port <= 0:
            raise ValueError("Given port number is not int or invalid")

        self._clients.append((address, port))

    def remove(self, address, port):
        """Remove a recipient

        Args:
            address (str): ip address of server
            port (int): port of server
        """

        for client in self._clients:
            if client[0] == address and client[1] == port:
                self._clients.remove(client)

                break

    def clear(self):
        """Clear list of receipts"""

        self._clients = []

    def send(self, message):
        """Sends an OSCBundle or OSCMessage to the servers.

        Args:
            message (OSCMessage, OSCBundle): a OSCMessage or OSCBundle to send
        """

        if not (isinstance(message, OSCMessage) or isinstance(message, OSCBundle)):
            raise ValueError("Given message is not a OSCMessage or OSCBundle")

        # create new socket if previously closed
        if self._closed:
            self._socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
            self._socket.setblocking(0)
            self._closed = False

        dgram = message.build().dgram

        for address in self._clients:
            self._socket.sendto(dgram, address)

    def close(self):
        """Close socket connection"""

        if not self._closed:
            self._socket.close()
            self._closed = True


class _UDPRequestHandler(socketserver.BaseRequestHandler):
    """Handles correct UDP messages for all types of server.

    Whether this will be run on its own thread, the server's or a whole new
    process depends on the server you instantiated, look at their documentation.

    This method is called after a basic sanity check was done on the datagram,
    basically whether this datagram looks like an osc message or bundle,
    if not the server won't even bother to call it and so no new
    threads/processes will be spawned.
    """

    def handle(self):
        """Handle UDP request and call handler callback"""

        data = self.request[0]
        callback = self.server.handle

        # Get OSC messages from all bundles or standalone message.
        try:
            packet = OSCPacket(data)
            now = calendar.timegm(time.gmtime())

            # If the message is to be handled later, then so be it.
            if packet.time > now:
                time.sleep(packet.time - now)

            callback(self.client_address, packet.message, packet.time)
        except OSCParseError:
            logging.warning("OSCParseError: Could not parse OSC packet")


class OSCServer(socketserver.UDPServer):
    """Superclass for different flavors of OSCServer

    You can change server logic by extending from both
    OSCServer and socketserver.ThreadingMixIn or socketserver.ForkingMixIn
    """

    def __init__(self, address='127.0.0.1', port=9000):
        """Initialize OSCServer class

        Args:
            address (string): string representation of ip address, for example: '127.0.0.1'
            port (int): port of server
        """

        super(OSCServer, self).__init__((address, port), _UDPRequestHandler)

    def verify_request(self, request, client_address):
        """Returns true if the data looks like a valid OSC UDP datagram.

        Args:
            request: A request data
            client_address: Client address
        Returns:
            True if request is valid
        """

        data = request[0]

        return OSCBundle.is_valid(data) or OSCMessage.is_valid(data)

    def server_bind(self):
        """Called by constructor to bind the socket."""
        self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.socket.bind(self.server_address)
        self.server_address = self.socket.getsockname()

    def handle(self, address, message, date):
        """Handle receiving of OSCMessage or OSCBundle

        Args:
            address: typle (host, port)
            message: OSCMessage or OSCBundle
            date: int number which represents time of message
        Raises:
            NotImplementedError if you don't override it
        """

        raise NotImplementedError("Re-implement this method")



# -*- coding: UTF-8 -*-
"""
    grailkit.bible_parse
    ~~~~~~~~~~~~~~~~~~~~

    Parse other bible formats into grail bible format.

    Bibles in OSIS format can be downloaded here:
    https://github.com/matt-cook/bible

    :copyright: (c) 2017-2019 by Oleksii Lytvyn.
    :license: MIT, see LICENSE for more details.
"""
import os
import xml.etree.ElementTree as ElementTree

from grailkit.dna import DNA
from grailkit.bible import BibleError


class Names(object):
    """This class holds book names and abbreviations
    to simplify parsing of bible formats
    """

    _books = [
        ["Genesis", "Genesis", ["Gen", "Ge", "Gn"]],
        ["Exodus", "Exodus", ["Exo", "Ex", "Exod"]],
        ["Leviticus", "Leviticus", ["Lev", "Le", "Lv"]],
        ["Numbers", "Numbers", ["Num", "Nu", "Nm", "Nb"]],
        ["Deuteronomy", "Deuteronomy", ["Deut", "Dt"]],
        ["Joshua", "Josh", ["Josh", "Jos", "Jsh"]],
        ["Judges", "Judges", ["Judg", "Jdg", "Jg", "Jdgs"]],
        ["Ruth", "Ruth", ["Rth", "Ru"]],
        ["1 Samuel", "First Samuel",
         ["1 Sam", "1 Sa", "1Samuel", "1S", "I Sa", "1 Sm", "1Sa", "I Sam", "1Sam", "I Samuel", "1st Samuel",
          "First Samuel"]],
        ["2 Samuel", "Second Samuel",
         ["2 Sam", "2 Sa", "2S", "II Sa", "2 Sm", "2Sa", "II Sam", "2Sam", "II Samuel", "2Samuel", "2nd Samuel",
          "Second Samuel"]],
        ["1 Kings", "First Kings",
         ["1 Kgs", "1 Ki", "1K", "I Kgs", "1Kgs", "I Ki", "1Ki", "I Kings", "1Kings", "1st Kgs", "1st Kings",
          "First Kings", "First Kgs", "1Kin"]],
        ["2 Kings", "Second Kings",
         ["2 Kgs", "2 Ki", "2K", "II Kgs", "2Kgs", "II Ki", "2Ki", "II Kings", "2Kings", "2nd Kgs", "2nd Kings",
          "Second Kings", "Second Kgs", "2Kin"]],
        ["1 Chronicles", "First Chronicles",
         ["1 Chron", "1 Ch", "I Ch", "1Ch", "1 Chr", "I Chr", "1Chr", "I Chron", "1Chron", "I Chronicles",
          "1Chronicles", "1st Chronicles", "First Chronicles"]],
        ["2 Chronicles", "Second Chronicles",
         ["2 Chron", "2 Ch", "II Ch", "2Ch", "II Chr", "2Chr", "II Chron", "2Chron", "II Chronicles", "2Chronicles",
          "2nd Chronicles", "Second Chronicles"]],
        ["Ezra", "Ezra", ["Ezra", "Ezr"]],
        ["Nehemiah", "Nehemiah", ["Neh", "Ne"]],
        ["Esther", "Esther", ["Add Esth", "Add Es", "Rest of Esther", "The Rest of Esther", "AEs", "AddEsth", "Est"]],
        ["Job", "Job", ["Job", "Jb"]],
        ["Psalms", "Psalms of Solomon", ["Ps Solomon", "Ps Sol", "Psalms Solomon", "PsSol"]],
        ["Proverbs", "Proverbs", ["Prov", "Pr", "Prv"]],
        ["Ecclus", "Ecclesiasticus", ["Sirach", "Sir", "Ecclesiasticus", "Ecclus"]],
        ["Song", "Song of Solomon", ["Song"]],
        ["Isaiah", "Isaiah", ["Isa"]],
        ["Jeremiah", "Letter of Jeremiah", ["Let Jer", "Let Jer", "LJe", "Ltr Jer"]],
        ["Lamentations", "Lamentations", ["Lam", "La"]],
        ["Ezekiel", "Ezekiel", ["Ezek", "Eze", "Ezk"]],
        ["Daniel", "Daniel", ["Dan", "Da", "Dn"]],
        ["Hosea", "Hosea", ["Hos", "Ho"]],
        ["Joel", "Joel", ["Joel", "Joe", "Jl"]],
        ["Amos", "Amos", ["Amos", "Am"]],
        ["Obadiah", "Obadiah", ["Obad", "Ob"]],
        ["Jonah", "Jonah", ["Jnh", "Jon"]],
        ["Nahum", "Nahum", ["Nah", "Na"]],
        ["Micah", "Micah", ["Micah", "Mic"]],
        ["Habakkuk", "Habakkuk", ["Hab", "Hab"]],
        ["Zephaniah", "Zephaniah", ["Zeph", "Zep", "Zp"]],
        ["Ephesians", "Ephesians", ["Ephes", "Eph"]],
        ["Haggai", "Haggai", ["Haggai", "Hag", "Hg"]],
        ["Zechariah", "Zechariah", ["Zech", "Zec", "Zc"]],
        ["Malachi", "Malachi", ["Mal", "Mal", "Ml"]],
        ["Matthew", "Matthew", ["Matt", "Mt"]],
        ["Mark", "Mark", ["Mrk", "Mk", "Mr"]],
        ["Luke", "Luke", ["Luk", "Lk"]],
        ["3 John", "Third John",
         ["3 John", "3 Jn", "III Jn", "3Jn", "III Jo", "3Jo", "III Joh", "3Joh", "III Jhn", "3 Jhn", "3Jhn", "III John",
          "3John", "3rd John", "Third John"]],
        ["Acts", "Acts", ["Acts", "Ac"]],
        ["Romans", "Romans", ["Rom", "Ro", "Rm"]],
        ["1 Corinthians", "First Corinthians",
         ["1 Cor", "1 Co", "I Co", "1Co", "I Cor", "1Cor", "I Corinthians", "1Corinthians", "1st Corinthians",
          "First Corinthians"]],
        ["2 Corinthians", "Second Corinthians",
         ["2 Cor", "2 Co", "II Co", "2Co", "II Cor", "2Cor", "II Corinthians", "2Corinthians", "2nd Corinthians",
          "Second Corinthians"]],
        ["Galatians", "Galatians", ["Gal", "Ga"]],
        ["Philippians", "Philippians", ["Philippians", "Phm"]],
        ["Colossians", "Colossians", ["Col", "Col"]],
        ["1 Thessalonians", "First Thessalonians",
         ["1 Thess", "1 Th", "I Th", "1Th", "I Thes", "1Thes", "I Thess", "1Thess", "I Thessalonians", "1Thessalonians",
          "1st Thessalonians", "First Thessalonians"]],
        ["2 Thessalonians", "Second Thessalonians",
         ["2 Thess", "2 Th", "II Th", "2Th", "II Thes", "2Thes", "II Thess", "2Thess", "II Thessalonians",
          "2Thessalonians", "2nd Thessalonians", "Second Thessalonians"]],
        ["1 Timothy", "First Timothy",
         ["1 Tim", "1 Ti", "I Ti", "1Ti", "I Tim", "1Tim", "I Timothy", "1Timothy", "1st Timothy", "First Timothy"]],
        ["2 Timothy", "Second Timothy",
         ["2 Tim", "2 Ti", "II Ti", "2Ti", "II Tim", "2Tim", "II Timothy", "2Timothy", "2nd Timothy",
          "Second Timothy"]],
        ["Titus", "Titus", ["Titus", "Tit"]],
        ["Philemon", "Philemon", ["Phil", "Phl", "Phlm"]],
        ["Hebrews", "Hebrews", ["Hebrews", "Heb"]],
        ["James", "James", ["James", "Jas", "Jm"]],
        ["1 Peter", "First Peter",
         ["1 Pet", "1 Pe", "I Pe", "1Pe", "I Pet", "1Pet", "I Pt", "1 Pt", "1Pt", "I Peter", "1Peter", "1st Peter",
          "First Peter"]],
        ["2 Peter", "Second Peter",
         ["2 Pet", "2 Pe", "II Pe", "2Pe", "II Pet", "2Pet", "II Pt", "2 Pt", "2Pt", "II Peter", "2Peter", "2nd Peter",
          "Second Peter"]],
        ["1 John", "First John",
         ["1 John", "1 Jn", "I Jn", "1Jn", "I Jo", "1Jo", "I Joh", "1Joh", "I Jhn", "1 Jhn", "1Jhn", "I John", "1John",
          "1st John", "First John"]],
        ["2 John", "Second John",
         ["2 John", "2 Jn", "II Jn", "2Jn", "II Jo", "2Jo", "II Joh", "2Joh", "II Jhn", "2 Jhn", "2Jhn", "II John",
          "2John", "2nd John", "Second John"]],
        ["3 John", "Third John",
         ["3 Jn", "III Jn", "3Jn", "III Jo", "3Jo", "III Joh", "3Joh", "III Jhn", "3 Jhn", "3Jhn", "III John", "3John",
          "3rd John"]],
        ["Jude", "Jude", ["Jude", "Jud"]],
        ["Revelation", "The Revelation", ["Rev", "Re"]]
    ]

    # Russian book names and abbreviations
    _books_ru = [
        ["Бытие", "Бытие", ["быт"]],
        ["Исход", "Исход", ["исх"]],
        ["Левит", "Левит", ["лев"]],
        ["Числа", "Числа", ["чис"]],
        ["Второзаконие", "Второзаконие", ["вт", "втор"]],

        ["Иисуса Навина", "Книга Иисуса Навина", ["Навина"]],
        ["Судей", "Книга Судей Израилевых", ["суд"]],
        ["Руфь", "Книга Руфи", ["ру", "руф"]],
        ["1 царств", "Первая книга царств", ["1цар", "цар", "1 цар"]],
        ["2 царств", "Вторая книга царств", ["2цар", "2 цар"]],
        ["3 царств", "Третья книга царств", ["3цар", "3 цар"]],
        ["4 царств", "Четвертая книга царств", ["4цар", "4 цар"]],
        ["1 паралипоменон", "Первая книга паралипоменон", ["пар", "1пар", "1 пар"]],
        ["2 паралипоменон", "Вторая книга паралипоменон", ["2пар", "2 пар"]],

        ["1 ездры", "Первая Книга Ездры", ["ездры", "езд"]],
        ["Неемии", "Книга Неемии", ["нем"]],
        ["Есфири", "Книга Есфири", ["есф"]],
        # ["2 ездры", "Вторая Книга Ездры", ["2езд"]],
        ["Иова", "Книга Иова", ["иов"]],

        ["Псалтирь", "Псалтирь", ["пс"]],
        ["Притчи", "Притчи Соломона", ["сол"]],
        ["Екклезиаста", "Книга Екклезиаста", ["ек", "еккл", "екл"]],
        ["Песнь песней соломона", "Песнь песней Соломона", ["песн"]],
        # ["Книга премудрости Соломона", "Книга премудрости Соломона", ["прем", "сол"]],
        # ["Книга премудрости Иисуса", "книга премудрости иисуса, сына сирахова", ["син"]],
        ["Исаии", "Книга пророка Исаии", ["ис", "исаии"]],
        ["Иеремии", "Книга пророка Иеремии", ["ир", "иерем"]],
        ["Плач Иеремии", "Плач Иеремии", ["ир", "иерем"]],
        # ["Послание Иеремии", "Послание Иеремии", ["пс иер"]],

        ["Иезекииля", "Книга пророка Иезекииля", ["из"]],
        ["Даниила", "Книга пророка Даниила", ["дан"]],
        ["Осии", "Книга пророка Осии", ["ос"]],
        ["Иоиля", "Книга пророка Иоиля", ["иол"]],
        ["Амоса", "Книга пророка Амоса", ["ам"]],
        ["Авдия", "Книга пророка Авдия", ["ав", "авд"]],
        ["Ионы", "Книга пророка Ионы", ["ион"]],
        # ["Варуха", "Книга пророка Варуха", ["вар"]],
        ["Наума", "Книга пророка Наума", ["наум"]],
        ["Михея", "Книга пророка Михея", ["мих"]],
        ["Аввакума", "Книга пророка Аввакума", ["авв"]],
        ["Софонии", "Книга пророка Софонии", ["соф"]],
        ["к Ефесянам", "Послание к Ефесянам святого апостола Павла", ["еф"]],
        ["Аггея", "Книга пророка Аггея", ["аг"]],
        ["Захарии", "Книга пророка Захарии", ["зах"]],
        ["Малахии", "Книга пророка Малахии", ["мал"]],
        # ["3 Ездры", "Третья книга Ездры", ["3езд"]],

        ["от Матфея", "От Матфея святое благовествование", ["мат", "мф"]],
        ["от Марка", "От Марка святое благовествование", ["мк"]],
        ["от Луки", "От Луки святое благовествование", ["лук", "лк"]],
        ["от Иоанна", "От Иоанна святое благовествование", ["ин"]],
        ["Деяния", "Деяния святых апостолов", ["деян"]],
        ["к Римлянам", "Послание к Римлянам святого апостола Павла", ["к римлянам", "рим"]],
        ["1 Коринфянам", "Первое послание к Коринфянам святого апостола Павла", ["1кор"]],
        ["2 Коринфянам", "Второе послание к Коринфянам святого апостола Павла", ["2кор"]],
        ["к Галатам", "Послание к Галатам святого апостола Павла", ["гал"]],
        ["к Филиппийцам", "Послание к Филиппийцам святого апостола Павла", ["филп"]],
        ["к Колоссянам", "Послание к Колоссянам святого апостола Павла", ["кол"]],
        ["1 Фессалоникийцам", "Первое послание к Фессалоникийцам (солунянам) святого апостола Павла", ["1фес"]],
        ["2 Фессалоникийцам", "Второе послание к Фессалоникийцам (солунянам) святого апостола Павла", ["2фес"]],
        ["1 Тимофею", "Первое послание к Тимофею святого апостола Павла", ["1тим"]],
        ["2 Тимофею", "Второе послание к Тимофею святого апостола Павла", ["2тим"]],
        ["к Титу", "Послание к Титу святого апостола Павла", ["тит"]],
        ["к Филимону", "Послание к Филимону святого апостола Павла", ["фил"]],
        ["к Евреям", "Послание к Евреям святого апостола Павла", ["евр", "евреям"]],
        ["Иакова", "Соборное послание святого апостола Иакова", ["иак"]],
        ["1 Петра", "Первое соборное послание святого апостола Петра", ["1пет"]],
        ["2 Петра", "Второе соборное послание святого апостола Петра", ["2пет"]],
        ["1 Иоанна", "Первое соборное послание святого апостола Иоанна", ["1ин"]],
        ["2 Иоанна", "Второе соборное послание святого апостола Иоанна", ["2ин"]],
        ["3 Иоанна", "Третье соборное послание святого апостола Иоанна", ["3ин"]],
        ["Иуды", "Соборное послание святого апостола Иуды", ["иуды", "ид"]],
        ["Откровение", "Откровение святого Иоанна Богослова", ["откр", "отк"]]
    ]

    # OSIS book names with reference too book description
    _osis_book_names = {
        "Gen": 0, "Exod": 1, "Lev": 2, "Num": 3, "Deut": 4,
        "Josh": 5, "Judg": 6, "Ruth": 7, "1Sam": 8, "2Sam": 9,
        "1Kgs": 10, "2Kgs": 11, "1Chr": 12, "2Chr": 13, "Ezra": 14,
        "Neh": 15, "Esth": 16, "Job": 17, "Ps": 18, "Prov": 19,
        "Eccl": 20, "Song": 21, "Isa": 22, "Jer": 23, "Lam": 24,
        "Ezek": 25, "Dan": 26, "Hos": 27, "Joel": 28, "Amos": 29,
        "Obad": 30, "Jonah": 31, "Nah": 32, "Mic": 33, "Hab": 34,
        "Zeph": 35, "Eph": 36, "Hag": 37, "Zech": 38, "Mal": 39,
        "Matt": 40, "Mark": 41, "Luke": 42, "John": 43, "Acts": 44,
        "Rom": 45, "1Cor": 46, "2Cor": 47, "Gal": 48, "Phil": 49,
        "Col": 50, "1Thess": 51, "2Thess": 52, "1Tim": 53, "2Tim": 54,
        "Titus": 55, "Phlm": 56, "Heb": 57, "Jas": 58, "1Pet": 59,
        "2Pet": 60, "1John": 61, "2John": 62, "3John": 63, "Jude": 64, "Rev": 65
    }

    @classmethod
    def osis(cls, osis_id, language='en'):
        """Get book info by OSIS id

        Args:
            osis_id (str): OSIS identifier
            language (str): language code
        """

        if osis_id in cls._osis_book_names:
            index = cls._osis_book_names[osis_id]

            # Join Russian and English names for Russian version
            if language == 'ru':
                ru = cls._books_ru[index]
                en = cls._books[index]

                return [ru[0], ru[1], ru[2] + en[2]]
            else:
                return cls._books[index]
        else:
            return None


class Parser(DNA):
    """Basic implementation of parser"""

    def __init__(self, file_in, file_out):
        """Parse `file_in` into grail file `file_out`

        Args:
            file_in (str): path to input file
            file_out (str): path to output file
        """

        # add new tables to basic grail format file
        self._db_create_query += """
            DROP TABLE IF EXISTS books;
            CREATE TABLE books(id INTEGER PRIMARY KEY AUTOINCREMENT, osisid TEXT, name TEXT, title TEXT, abbr TEXT );

            DROP TABLE IF EXISTS verses;
            CREATE TABLE verses( osisid TEXT, book INT, chapter INT, verse INT, text TEXT );
            """

        super(Parser, self).__init__(file_out, create=True)

        if not os.path.isfile(file_in):
            raise BibleError("Could not open file. File %s not found." % (file_in,))

        # read input file and write to output file
        self.parse_file(file_in)

    def set_property(self, key, value):
        """Set bible property

        Args:
            key (str): property key
            value (str): property value
        """

        self._set(0, key, value, force_type=DNA.ARG_STRING)

    def get_property(self, key, default=""):
        """Get bible property

        Args:
            key (str): property name
            default (str): default value if property not exists

        Returns:
            bible property
        """

        return self._get(0, key, default=default)

    def write_verse(self, osis_id, book_id, chapter, verse, text):
        """Add verse to file

        Args:
            osis_id (str): OSIS identifier
            book_id (int): book number
            chapter (int): chapter number
            verse (int): verse number
            text (str): text of verse
        """

        self._db.execute("INSERT INTO verses VALUES(?, ?, ?, ?, ?)",
                         (osis_id, book_id, chapter, verse, text))

    def write_book(self, book_id, osis_id, name, title, abbr):
        """Add book to file

        Args:
            book_id (int): book number
            osis_id (str): OSIS identifier
            name (str): short book name
            title (str): full name of book
            abbr (str): string with abbreviations of book name separated by comma
        """

        self._db.execute("INSERT INTO books VALUES(?, ?, ?, ?, ?)",
                         (book_id, osis_id, name, title, abbr))

    def parse_file(self, file_in):
        """Abstract method. Implement this method in sub class.

        Args:
            file_in (str): path to file that will be parsed
        """

        raise NotImplementedError()


class OSISParser(Parser):
    """Parse OSIS bible format file."""

    def __init__(self, file_in, file_out):
        """Parse a OSIS file into grail file format bible

        Args:
            file_in (str): input file
            file_out (str): output file
        """

        self._osis_language = 'en'

        super(OSISParser, self).__init__(file_in, file_out)

    def parse_file(self, file_in):
        """Parse input file

        Args:
            file_in (str): path to file to be parsed
        """

        tree = ElementTree.parse(file_in)
        root = tree.getroot()
        book = 0

        # parse translation info first
        for book_node in root[0]:
            tag = book_node.tag.split('}')[-1]

            if tag == 'header':
                self._parse_header(book_node)

        self._osis_language = self.get_property('language', default='en')

        # find and parse all books & verses
        for book_node in root[0]:
            tag = book_node.tag.split('}')[-1]

            if tag == 'div' and book_node.attrib['type'] and book_node.attrib['type'] == 'book':
                book += 1
                book_osisid = book_node.attrib['osisID']
                book_info = self._get_book_info(book_osisid)

                self.write_book(book, book_osisid, book_info[0], book_info[1], book_info[2])

                for chapter_node in book_node:
                    for verse_node in chapter_node:
                        if verse_node.tag.split('}')[-1] == 'verse':
                            osis_id = verse_node.attrib['osisID']
                            verse_id = osis_id.split('.')

                            self.write_verse(osis_id, book, int(verse_id[1]), int(verse_id[2]), verse_node.text)

        self.set_property('version', '1.0')

    def _parse_header(self, node):
        """Parse osis header node

        Args:
            node: header xml node
        """

        for headnode in node:
            if headnode.tag.split('}')[-1] == 'work':
                self._parse_header_property(headnode)

    def _parse_header_property(self, node):
        """Parse header property

        Args:
            node: node of header
        """
        for prop in node:
            name = prop.tag.split('}')[-1]
            ids = {
                'date': '',
                'title': '',
                'subject': '',
                'language': '',
                'publisher': '',
                'rights': 'copyright',
                'identifier': '',
                'description': ''
                }

            # check if this property is valid
            if name in ids:
                key = ids[name] if ids[name] != "" else name
                value = prop.text if prop.text else ""

                self.set_property(key, value)

    def _get_book_info(self, osis_id):
        """Get info of book using OSIS identifier

        Args:
            osis_id (str): OSIS identifier
        Returns:
            array that consists of three items
            first item is short name of book
            second item is full name of book
            and third is abbreviations separated by comma
        """

        book = Names.osis(osis_id, self._osis_language)

        return [book[0], book[1], ", ".join(book[2])] if book else [osis_id, osis_id, osis_id]



# -*- coding: UTF-8 -*-
"""
    grailkit.core
    ~~~~~~~~~~~~~

    Core types and widely used components

    :copyright: (c) 2017-2019 by Oleksii Lytvyn.
    :license: MIT, see LICENSE for more details.
"""
import weakref
import logging

from grailkit.util import object_type

logging.getLogger(__name__).addHandler(logging.NullHandler())


class Signal(object):
    """Callback mechanism for DNA objects
    This class uses weak references to callbacks, so methods can be deleted
    if no references exists"""

    def __init__(self, *args):
        """Create signal

        Args:
            *args: list of types, used as template of arguments
        """

        self._args = [object_type(x) for x in args]
        self._fns = {}
        self._flush_keys = []

    def __len__(self):
        """Returns number of connected slots"""

        return len(self._fns)

    def __bool__(self):
        """Returns True and prevent from converting to False when number of slots is 0"""

        return True

    @property
    def template(self):
        """Returns string template of types of signal"""

        return ", ".join("<%s>" % str(x.__name__) for x in self._args)

    def connect(self, fn, name=None):
        """Add function to list of callbacks

        Args:
            fn (callable): function to call on emit
            name (str): give name to slot
        """

        if not callable(fn):
            raise ValueError("Given object is not callable")

        ref = self._wrap(fn)

        if not name:
            name = len(self._fns)

        self._fns[name] = ref

    def disconnect(self, fn):
        """Remove function from list, if it previously added to it

        Args:
            fn (callable): function to remove
        """

        found_key = None

        for key, value in self._fns.items():
            if value[0] and fn == value[1]():
                found_key = key

                break

        if found_key:
            del self._fns[found_key]

    def emit(self, *args, name=False, **kwargs):
        """Emit signal
        If `name` argument was given, only slot with this name will be called
        otherwise all slots will be called

        Args:
            *args: arguments to pass to callbacks
            name (str): give name of slot to be called
            **kwargs: keyword arguments to pass to callbacks
        """

        if name and name in self._fns:
            return self._call(name, *args, **kwargs)

        for key in self._fns:
            self._call(key, *args, **kwargs)

        # remove dead references
        self._flush()

    def _call(self, name, *args, **kwargs):
        """Call method, if reference is dead then flush

        Args:
            name (str): slot name
            *args (list): list of argument to be passed
            **kwargs (dict): dict of arguments to be passed
        """

        ref = self._fns[name]
        obj = ref[0]
        fun = ref[1]

        try:
            # bound
            if callable(obj) and obj() and callable(fun):
                callback = getattr(obj(), fun().__name__)
                callback(*args, **kwargs)
            # unbound
            elif obj is None and fun:
                fun(*args, **kwargs)
            # non exists
            else:
                self._flush_keys.append(name)
        # Error caused by callback
        except RuntimeError as e:
            logging.warning("Slot %s removed because of exception raised.\n "
                            "Original exception was: %s" %
                            (str(fun), str(e)))

            self._flush_keys.append(name)

    def _flush(self):
        """Delete dead references"""

        for key in set(self._flush_keys):
            if key in self._fns:
                del self._fns[key]

        self._flush_keys = []

    @classmethod
    def _wrap(cls, fn):
        """Returns typle with parent object and method

        Args:
            fn (callable): callable to wrap with weakref
        """

        if hasattr(fn, '__self__') and hasattr(fn, '__func__'):
            return weakref.ref(fn.__self__), weakref.ref(fn.__func__)
        else:
            return None, fn


class Signalable(object):
    """Like a Signal but with messages and bundles"""

    def __init__(self):

        self.__slots = {}
        self.__bundle_slots = Signal()

    def __bool__(self):
        """Returns True"""

        return True

    def __len__(self):
        """Returns number of registered callbacks"""

        return self.callbacks_length

    @property
    def callbacks_length(self):
        """Returns number of registered callbacks"""

        return sum(len(v) for k, v in self.__slots.items()) + len(self.__bundle_slots)

    def connect(self, message, fn):
        """Connect listener `fn` to slot `message`

        Args:
            message (str): slot name
            fn (callable): function to call
        Raises:
            ValueError if at least one of arguments is not supported
        """

        if not isinstance(message, str):
            raise ValueError("Can't connect to slot '%s', given value is not of type string" % message)

        if not callable(fn):
            raise ValueError("Given function is not callable.")

        if message not in self.__slots:
            self.__slots[message] = Signal()

        self.__slots[message].connect(fn)

    def disconnect(self, message, fn):
        """Disconnect listener from slot

        Args:
            message (str): slot name
            fn (callable): function to call
        """

        if message in self.__slots:
            self.__slots[message].disconnect(fn)

    def emit(self, message, *args):
        """Trigger all listeners of message

        Args:
            message (str): slot name
            *args: list of arguments
        """
        if message in self.__slots:
            self.__slots[message].emit(*args)

    def connect_bundle(self, fn):
        """Connect a bundle listener

        Args:
            fn (callable): callback listener
        """

        self.__bundle_slots.connect(fn)

    def disconnect_bundle(self, fn):
        """Remove bundle listener

        Args:
            fn (callable): callback listener
        """

        self.__bundle_slots.disconnect(fn)

    def emit_bundle(self, bundle):
        """Emit bundle of messages

        Args:
            bundle (list): bundle of messages
        """

        self.__bundle_slots.emit(bundle)



# -*- coding: UTF-8 -*-
"""
    grailkit.bible
    ~~~~~~~~~~~~~~

    Implements access to bible files.

    Grail bible format consists of standard grail DNA format plus
    two additional tables: `books` and `verses`

    SQL structure:

    .. code:: sql

        CREATE TABLE books(id INTEGER PRIMARY KEY AUTOINCREMENT, osisid TEXT, name TEXT, title TEXT, abbr TEXT );
        CREATE TABLE verses( osisid TEXT, book INT, chapter INT, verse INT, text TEXT );

    :copyright: (c) 2017-2019 by Oleksii Lytvyn.
    :license: MIT, see LICENSE for more details.
"""
import os
import re
import glob
import json
import sqlite3
import logging

from grailkit import PATH_SHARED
from grailkit.util import copy_file, default_key, file_exists
from grailkit.dna import DNA


def verse_factory(cursor, row):
    """Parse sqlite row into Verse object

    Args:
        cursor (sqlite3.Cursor): sqlite3 cursor
        row (sqlite3.Row): row data

    Returns:
        Verse object parsed from sqlite row
    """

    return Verse.from_sqlite(row)


def book_factory(cursor, row):
    """Parse sqlite row into Book object

    Args:
        cursor (sqlite3.Cursor): sqlite3 cursor
        row (sqlite3.Row): row data

    Returns:
        Book object parsed from sqlite row
    """

    return Book.from_sqlite(row)


class BibleError(Exception):
    """Base error thrown when a bible could to be read."""

    pass


class Verse:
    """Representation of Bible verse"""

    def __init__(self):
        """Create verse"""

        self._osisid = ""
        self._book = ""
        self._book_id = 1
        self._chapter = 1
        self._verse = 1
        self._text = ""

    @property
    def type(self):
        """DNA type (int)"""

        return DNA.TYPE_VERSE

    @property
    def book(self):
        """Book name (str)"""

        return self._book

    @property
    def book_id(self):
        """Book id (int)"""

        return self._book_id

    @property
    def chapter(self):
        """Chapter number (int)"""

        return self._chapter

    @property
    def verse(self):
        """Verse number (int)"""

        return self._verse

    @property
    def reference(self):
        """Returns complete reference string

        Examples:
            Genesis 1:1
            1 Corinthians 2:12
        """

        return "%s %d:%d" % (self._book, self._chapter, self._verse)

    @property
    def text(self):
        """Text of verse"""

        return self._text

    def parse(self, row):
        """Parse sqlite row into verse"""

        self._book = row[5]
        self._book_id = row[1]
        self._chapter = row[2]
        self._verse = row[3]
        self._text = row[4]
        self._osisid = row[0]

    @staticmethod
    def from_sqlite(row):
        """Parse sqlite row and return Verse

        Args:
            row: sqlite3 row

        Returns:
            Verse parsed from given sqlite row
        """

        verse = Verse()
        verse.parse(row)

        return verse


class Book:
    """Representation of bible book"""

    def __init__(self):
        """Create book"""

        self._id = 0
        self._abbr = ""
        self._name = ""
        self._title = ""
        self._osisid = ""

    @property
    def type(self):
        """DNA type (int)"""

        return DNA.TYPE_BOOK

    @property
    def id(self):
        """Book abbreviations (str)"""

        return self._id

    @property
    def abbr(self):
        """Book abbreviations (str)"""

        return self._abbr

    @property
    def name(self):
        """Name of book (str)"""

        return self._name

    @property
    def title(self):
        """Full name of book, it might be bigger than name"""

        return self._title

    @property
    def osisid(self):
        """OSIS identifier, can be used for cross-referencing"""

        return self._osisid

    def parse(self, row):
        """Parse sqlite row and fill Book"""

        self._id = row[0]
        self._abbr = row[4]
        self._name = row[2]
        self._title = row[3]
        self._osisid = row[1]

    @staticmethod
    def from_sqlite(row):
        """Parse sqlite row and return Book

        Args:
            row: sqlite row

        Returns:
            Book
        """

        book = Book()
        book.parse(row)

        return book


class BibleInfo:
    """Read only representation of bible file"""

    def __init__(self):
        """Create a object"""

        self._file = ""
        self._date = ""
        self._title = ""
        self._subject = ""
        self._language = ""
        self._publisher = ""
        self._copyright = ""
        self._identifier = ""
        self._description = ""

        self._version = 1

    @property
    def file(self):
        """File location"""

        return self._file

    @property
    def date(self):
        """Date of publication"""

        return self._date

    @property
    def title(self):
        """Bible title"""

        return self._title

    @property
    def subject(self):
        """Subject of a bible"""

        return self._subject

    @property
    def language(self):
        """Language of bible"""

        return self._language

    @property
    def publisher(self):
        """Publisher information"""

        return self._publisher

    @property
    def copyright(self):
        """Copyright information"""

        return self._copyright

    @property
    def identifier(self):
        """Bible identifier, must be unique"""

        return self._identifier

    @property
    def description(self):
        """A little description of Bible"""

        return self._description

    @property
    def version(self):
        """Schema version number"""

        return self._version

    @staticmethod
    def from_json(data):
        """Fill properties from json string

        Args:
            data (dict): parsed json object
        Returns:
            BibleInfo object
        """

        info = BibleInfo()

        info._file = default_key(data, 'file', '')
        info._date = default_key(data, 'date')
        info._title = default_key(data, 'title')
        info._subject = default_key(data, 'subject')
        info._language = default_key(data, 'language')
        info._publisher = default_key(data, 'publisher')
        info._copyright = default_key(data, 'copyright')
        info._identifier = default_key(data, 'identifier')
        info._description = default_key(data, 'description')

        return info


class Bible(DNA):
    """Representation of grail bible file.
    This class gives you read only access to file
    """

    # file extension
    _file_extension = ".grail-bible"

    def __init__(self, file_path):
        """Read grail bible file into Bible class

        Args:
            file_path (str): file location

        Raises:
            DNAError if file does not exists
        """

        super(Bible, self).__init__(file_path, create=False)

        # read bible info
        self._date = self._get(0, "date", default="")
        self._title = self._get(0, "title", default="Untitled")
        self._subject = self._get(0, "subject", default="")
        self._language = self._get(0, "language", default="unknown")
        self._publisher = self._get(0, "publisher", default="Unknown Publisher")
        self._copyright = self._get(0, "copyright", default="copyright information unavailable")
        self._identifier = self._get(0, "identifier", default="NONE")
        self._description = self._get(0, "description", default="")

        self._version = self._get(0, "version", default=1)

    @property
    def date(self):
        """Date of publication"""

        return self._date

    @property
    def title(self):
        """Bible title"""

        return self._title

    @property
    def subject(self):
        """Subject of a bible"""

        return self._subject

    @property
    def language(self):
        """Language of bible"""

        return self._language

    @property
    def publisher(self):
        """Publisher information"""

        return self._publisher

    @property
    def copyright(self):
        """Copyright information"""

        return self._copyright

    @property
    def identifier(self):
        """Bible identifier, must be unique"""

        return self._identifier

    @property
    def description(self):
        """A little description of Bible"""

        return self._description

    @property
    def version(self):
        """Schema version number"""

        return self._version

    def books(self):
        """Returns list of all books"""

        return self._db.all("""SELECT 
            `books`.`id`,
            `books`.`osisid`,
            `books`.`name`,
            `books`.`title`,
            `books`.`abbr`
            FROM books""", factory=book_factory)

    def book(self, book):
        """Returns single book

        Args:
            book (int): book id
        """

        return self._db.get("""SELECT
            `books`.`id`,
            `books`.`osisid`,
            `books`.`name`,
            `books`.`title`,
            `books`.`abbr` FROM books WHERE id = ?""", (book,), factory=book_factory)

    def chapter(self, book, chapter):
        """Returns all verses in chapter

        Args:
            book (int): book id
            chapter (int): chapter id
        """

        return self._db.all("""SELECT 
                `verses`.`osisid`,
                `verses`.`book`,
                `verses`.`chapter`,
                `verses`.`verse`,
                `verses`.`text`,
                `books`.`name` as book_name
            FROM verses
            LEFT JOIN `books` ON `verses`.`book` = `books`.`id`
            WHERE `verses`.`book` = ? AND `verses`.`chapter` = ?
            ORDER BY `verses`.`verse` ASC""", (book, chapter), factory=verse_factory)

    def verse(self, book, chapter, verse):
        """Returns single verse

        Args:
            book (int): book id
            chapter (int): chapter id
            verse (int): verse number
        """

        return self._db.get("""SELECT 
                                `verses`.`osisid`,
                                `verses`.`book`,
                                `verses`.`chapter`,
                                `verses`.`verse`,
                                `verses`.`text`,
                                `books`.`name` as book_name
                            FROM verses
                            LEFT JOIN `books` ON `verses`.`book` = `books`.`id`
                            WHERE `verses`.`book` = ? AND `verses`.`chapter` = ? AND `verses`.`verse` = ?""",
                            (book, chapter, verse), factory=verse_factory)

    def count_verses(self, book, chapter):
        """Returns number of verses in chapter

        Args:
            book (int): book id
            chapter (int): chapter id
        """

        return self._db.get("SELECT COUNT(*) as count FROM verses WHERE book = ? AND chapter = ?",
                            (book, chapter))["count"]

    def count_chapters(self, book):
        """Returns number of chapters in book

        Args:
            book (int): book id
        """

        return self._db.get("SELECT COUNT(*) as count FROM verses WHERE book = ? AND verse = 1",
                            (book,))["count"]

    def match_book(self, keyword):
        """Find books by keyword

        Args:
            keyword (str): search phrase
        """

        keyword = "%" + keyword + "%"

        return self._db.all("""SELECT
                            `books`.`id`,
                            `books`.`osisid`,
                            `books`.`name`,
                            `books`.`title`,
                            `books`.`abbr`
                            FROM books
                            WHERE
                             lowercase(`title`) LIKE lowercase( ? )
                             OR lowercase(`short`) LIKE lowercase( ? )
                             OR lowercase(`full`) LIKE lowercase( ? )
                            """, (keyword, keyword, keyword), factory=book_factory)

    def match_reference(self, keyword, limit=3):
        """find verse by keyword

        Args:
            keyword (str): keywords
            limit (int): limit results
        """

        # default values
        chapter = 1
        verse = 1
        keyword = keyword.lstrip().rstrip().lower()

        # match book
        match = re.search(r'([0-9]+)$', keyword)
        # match book and chapter
        match_chapter = re.search(r'([0-9]+)([\D])([0-9]+)$', keyword)
        # match book, chapter and verse
        match_verse = re.search(r'([0-9]+)([\D])([0-9]+)-([0-9]+)$', keyword)

        if match_verse:
            chapter = match_verse.group(1)
            verse = match_verse.group(3)
            keyword = re.sub(r'([0-9]+)([\D])([0-9]+)-([0-9]+)$', '', keyword)
        elif match_chapter:
            chapter = match_chapter.group(1)
            verse = match_chapter.group(3)
            keyword = re.sub(r'([0-9]+)([\D])([0-9]+)$', '', keyword)
        elif match:
            chapter = match.group(1)
            keyword = re.sub(r'([0-9]+)$', '', keyword)

        keyword = "%" + keyword.lstrip().rstrip() + "%"
        chapter = int(chapter)
        verse = int(verse)

        return self._db.all("""SELECT
                                `verses`.`osisid`,
                                `verses`.`book`,
                                `verses`.`chapter`,
                                `verses`.`verse`,
                                `verses`.`text`,
                                `books`.`name` as book_name
                            FROM verses
                            LEFT JOIN `books` ON `verses`.`book` = `books`.`id`
                            WHERE
                                (lowercase(`books`.`title`) LIKE ?
                                OR lowercase(`books`.`name`) LIKE ?
                                OR lowercase(`books`.`abbr`) LIKE ?)
                                AND `verses`.`chapter` = ?
                                AND `verses`.`verse` = ?
                            LIMIT ?""",
                            (keyword, keyword, keyword, chapter, verse, limit), factory=verse_factory)

    def match_text(self, text, limit=3):
        """Search for text occurrences

        Args:
            text (str): search string
            limit (int): limit results
        """

        keyword = "%" + text.lstrip().rstrip().lower() + "%"

        return self._db.all("""SELECT 
                                `verses`.`osisid`,
                                `verses`.`book`,
                                `verses`.`chapter`,
                                `verses`.`verse`,
                                `verses`.`text`,
                                `books`.`name` as book_name
                               FROM verses 
                               LEFT JOIN `books` ON `verses`.`book` = `books`.`id`
                               WHERE lowercase(`verses`.`text`) LIKE ? LIMIT ?""",
                            (keyword, limit), factory=verse_factory)

    def json_info(self):
        """Create json information string"""

        data = {
            "file": self._location,
            "date": self._date,
            "title": self._title,
            "subject": self._subject,
            "language": self._language,
            "publisher": self._publisher,
            "copyright": self._copyright,
            "identifier": self._identifier,
            "description": self._description
            }

        return json.dumps(data)


class BibleHostError(Exception):
    """BibleHost errors"""

    pass


class BibleHost:
    """Manage all installed bibles."""

    # location of shared folder
    _location = os.path.join(PATH_SHARED, "bibles/")

    # list of available bibles
    _list = {}
    _list_refs = {}

    @classmethod
    def setup(cls):
        """Gather information about installed bibles.
        Call this method before use of BibleHost."""

        cls._list = {}

        for f in glob.glob(cls._location + "*.json"):
            file = open(f, "r")
            info = BibleInfo.from_json(json.load(file))
            file.close()

            cls._list[info.identifier] = info

    @classmethod
    def list(cls):
        """List all installed bibles

        Returns:
            list of BibleInfo objects
        """

        return cls._list

    @classmethod
    def info(cls, bible_id):
        """Get a bible info object

        Args:
            bible_id (str): bible identifier
        Returns:
            Bible object if exists otherwise None
        """

        if bible_id in cls._list:
            return cls._list[bible_id]

        return None

    @classmethod
    def get(cls, bible_id):
        """Get a Bible by identifier

        Args:
            bible_id (str): bible identifier string
        Returns:
            Bible object if bible exists under given identifier
        """

        bible = cls.info(bible_id)

        if bible:
            cls._list_refs[bible_id] = Bible(bible.file)

            return cls._list_refs[bible_id]

        return None

    @classmethod
    def install(cls, file_path, replace=False):
        """Install bible from file, grail-bible format only

        Args:
            file_path (str): path to bible file
            replace (bool): if bible already installed replace it by another version
        Returns:
            True if installed or False if bible
        Raises:
            BibleHostError raised if bible identifier already exists
        """

        if not cls.verify(file_path):
            raise BibleHostError("Unable to install bible from file \"%s\" due to file corruption." % file_path)

        try:
            bible = Bible(file_path)
            bible.close()
            bible_path = os.path.join(cls._location, bible.identifier + '.grail-bible')
        except (BibleError, sqlite3.OperationalError):
            raise BibleHostError("Unable to install bible from file \"%s\"."
                                 "File may be corrupted or in another format." % file_path)

        if cls.info(bible.identifier) and not replace:
            raise BibleHostError("Bible %s (%s) already installed" % (bible.title, bible.identifier,))

        # just copy file to new location
        copy_file(file_path, bible_path)

        installed_bible = Bible(bible_path)
        # track bible descriptor
        cls._list_refs[installed_bible.identifier] = installed_bible

        # create a description file
        cls._create_descriptor(installed_bible)
        cls.setup()

        return True

    @classmethod
    def uninstall(cls, bible_id):
        """Uninstall bible by id

        Args:
            bible_id (str): bible identifier
        """

        if bible_id in cls._list:
            if bible_id in cls._list_refs:
                cls._list_refs[bible_id].close()
                del cls._list_refs[bible_id]

            del cls._list[bible_id]

        try:
            os.remove(os.path.join(cls._location, bible_id + ".json"))
            os.remove(os.path.join(cls._location, bible_id + ".grail-bible"))
        except Exception as e:
            logging.warning("Bible unistalled with error: %s" % e)

    @classmethod
    def verify(cls, file_path):
        """Check file to be valid grail-bible file

        Args:
            file_path (str): path to bible file
        Returns:
            True if bible file is valid
        """

        if not file_exists(file_path):
            return False

        return True

    @classmethod
    def _create_descriptor(cls, bible):
        """Create a description file for a bible

        Args:
            bible (Bible): bible object
        """

        data = bible.json_info()

        file = open(os.path.join(cls._location, bible.identifier + ".json"), "w")
        file.write(data)
        file.close()

        bible_info = BibleInfo()
        bible_info.from_json(json.loads(data))

        cls._list[bible_info.identifier] = bible_info



# -*- coding: UTF-8 -*-
"""
    grailkit.db
    ~~~~~~~~~~~

    Simplified interface to SQLite database

    :copyright: (c) 2017-2019 by Oleksii Lytvyn.
    :license: MIT, see LICENSE for more details.
"""
import os
import re
import sqlite3
import logging

from grailkit import util


logging.getLogger(__name__).addHandler(logging.NullHandler())


def create_factory(object_def, cursor, row):
    """Create object factory

    Args:
        object_def (callable): callable object
        cursor (sqlite3.Cursor): database cursor
        row (sqlite3.Row): database row object
    Returns:
        instance of `object_def`
    """

    if not object_def or not callable(object_def):
        raise DataBaseError("Can't create factory with given object. Object is not callable or not exists.")

    return object_def(row, cursor)


class DataBaseError(Exception):
    """Base class for DataBase Errors"""

    pass


class DataBase:
    """SQLite database wrapper"""

    # sqlite connection handler
    _connection = None

    def __init__(self, file_path, file_copy=False, query="", create=False):
        """Create SQLite database wrapper.
        Also define custom functions sql `lowercase` an `search_strip`.

        Args:
            file_path (str): database file path
            file_copy (str): copy file if file_path not exists
            query (str): execute query if file_path not exists
            create (bool): create database file or not
        """

        directory = os.path.dirname(os.path.realpath(file_path))
        execute_query = False

        if not create and not util.file_exists(file_path):
            raise DataBaseError("Database file not exists. Unable to open sqlite file @ %s." % file_path)

        if not os.path.exists(directory):
            os.makedirs(directory)

        if not os.path.isfile(file_path):
            if file_copy:
                util.copy_file(file_copy, file_path)
            else:
                file_handle = open(file_path, 'w+')
                file_handle.close()
                execute_query = True

        self._connection = sqlite3.connect(file_path)
        self._connection.row_factory = sqlite3.Row
        self._location = file_path

        def lowercase(char):
            """Lover string

            Args:
                char (str): string to process
            """
            return char.lower()

        def search_strip(char):
            """Prepare string for search

            Args:
                char (str): string to process
            """
            char = re.sub(r'[\[_\].\-,!()\"\':;]', '', char)
            char = re.sub('[\s+]', '', char)

            return char.lower()

        self._connection.create_function("lowercase", 1, lowercase)
        self._connection.create_function("search_strip", 1, search_strip)

        if execute_query and query:
            self.cursor.executescript(query)

        DataBaseHost.add(self)

    @property
    def connection(self):
        """Returns sqlite3 connection object"""

        return self._connection

    @property
    def cursor(self):
        """Returns sqlite3 connection cursor"""

        return self._connection.cursor()

    @property
    def location(self):
        """Returns location of database file"""

        return self._location

    def get(self, query, data=tuple(), factory=None):
        """Execute query and return first record

        Args:
            query (str): SQL query string
            data (tuple): tuple of data
            factory (callable): sqlite row factory
        Returns:
            first row
        """

        if factory:
            self.set_factory(factory)

        cursor = self.cursor
        cursor.execute(query, data)

        result = cursor.fetchone()
        self.set_factory()

        return result

    def all(self, query, data=tuple(), factory=None):
        """Execute query and return all records

        Args:
            query (str): SQL query string
            data (tuple): tuple of data
            factory (callable): sqlite row factory for this query only
        Returns:
            list of fetched rows
        """

        if factory:
            self.set_factory(factory)

        cursor = self.cursor
        cursor.execute(query, data)

        result = cursor.fetchall()
        self.set_factory()

        return result

    def execute(self, query, data=tuple()):
        """Execute many sql queries at once

        Args:
            query (str): SQL query string
            data (tuple): tuple of data
        """

        cursor = self.cursor
        cursor.execute(query, data)

    def set_factory(self, factory=sqlite3.Row):
        """Set sqlite row factory function.
        If you call `set_factory` without arguments default factory will be used

        Example:
            def string_factory(cursor, row):
                return [str(value) for value in row]

        Args:
            factory (callable): factory object
        """

        self._connection.row_factory = factory

    def copy(self, file_path, create=False):
        """Copy database to new file location

        Args:
            file_path (str): path to new file
            create (bool): create file if not exists
        """

        directory = os.path.dirname(os.path.realpath(file_path))

        if not file_path:
            raise ValueError('Path to the file is invalid')

        if not create and not util.file_exists(file_path):
            raise DataBaseError('Unable to copy database, file %s not exists.' % file_path)

        if not os.path.exists(directory):
            os.makedirs(directory)

        if not os.path.isfile(file_path):
            file_handle = open(file_path, 'w+')
            file_handle.close()

        db = sqlite3.connect(file_path)
        query = "".join(line for line in self._connection.iterdump())

        db.executescript(query)
        db.commit()
        db.close()

    def commit(self):
        """Commit changes to database"""

        self._connection.commit()

    def close(self):
        """Commit changes to database and close connection"""

        try:
            self._connection.commit()
        except sqlite3.ProgrammingError:
            logging.info("Unable to commit into %s, connection was closed" % (self._location, ))

        # close if this connection is not used by others
        if not DataBaseHost.has(self._location):
            self._connection.close()


class DataBaseHost:
    """Host all sqlite database connections"""

    # list of all connected databases
    _list = {}

    @classmethod
    def get(cls, file_path, file_copy=False, query="", create=True):
        """Get DataBase object

        Args:
            file_path (str): path to database file
            file_copy (str): copy file from `file_copy` if `file_path` not exists
            query (str): execute query if file `file_path` not exists
            create (bool): create database file or not
        Returns:
            DataBase object if opened or opens database and returns it.
        """

        file_path = os.path.abspath(file_path)

        if file_path in DataBaseHost._list:
            db = cls._list[file_path]
        else:
            db = DataBase(file_path, file_copy, query, create)
            cls._list[file_path] = db

        return db

    @classmethod
    def add(cls, db_ref):
        """Add DataBase object to list if not exists

        Args:
            db_ref (DataBase): reference to database error
        Raises:
            DataBaseError: If wrong object were passed
        """

        if not db_ref or not isinstance(db_ref, DataBase):
            raise DataBaseError("DataBase object doesn't exists or it's not an instance of DataBase class.")

        file_path = os.path.abspath(db_ref.location)

        if file_path not in cls._list:
            cls._list[db_ref.location] = db_ref

    @classmethod
    def has(cls, file_path):
        """Returns True if `file_path` in list of opened connections

        Args:
            file_path (str): file location
        """

        return os.path.abspath(file_path) in cls._list

    @staticmethod
    def close():
        """Close all connections

        Returns:
            bool: True if all connections closed successfully
        """

        for key in DataBaseHost._list:
            DataBaseHost._list[key].close()

        return True



# -*- coding: UTF-8 -*-
"""
    grailkit.dna
    ~~~~~~~~~~~~

    Implementation of grail file format.
    This file format used for all internal grail data structures.

    :copyright: (c) 2017-2019 by Oleksii Lytvyn.
    :license: MIT, see LICENSE for more details.
"""
import os
import json
import sqlite3 as lite

from grailkit.core import Signal
from grailkit.db import DataBaseHost, DataBaseError
from grailkit.util import millis_now, default_key

# todo: Add ability to change nodes without executing SQL queries, as each change generates events in current model


class DNAError(DataBaseError):
    """Base class for DNA errors"""

    pass


class DNAEntity:
    """DNA entity definition class

    Each entity must have following fields:
        id (int): identification number of entity;
        type (int): type of entity;
        name (str): name or label of entity;
        parent (int): parent entity identifier;
        search (str): string field that contains info for search;
        content (json): content of entity, can be uniform;
        created (int): time when entity is created;
        modified (int): time whe entity was last time modified;
        index (int): index of entity, in case of ordered lists of entities;
    """

    def __init__(self, parent):
        """DNA entity representation.
        This class provides methods and properties for easier management of entities an their properties

        Args:
            parent (DNA): parent DNA object
        """

        if not isinstance(parent, DNA):
            raise DNAError("DNAEntity can't be created without parent DNA. Given %s" % str(parent))

        self._id = 0
        self._type = DNA.TYPE_ABSTRACT
        self._name = ""
        self._parent = 0
        self._search = ""
        self._content = None
        self._created = millis_now()
        self._modified = 0
        self._index = 0
        self._dna = parent

    def __len__(self):
        """Returns number of properties"""

        return len(self.properties())

    def __bool__(self):
        """Boolean representation"""

        return True

    @property
    def id(self):
        """Returns entity identifier number"""

        return self._id

    @property
    def dna(self):
        """Returns parent DNA object, where this entity belongs"""

        return self._dna

    @property
    def parent(self):
        """Returns parent entity"""

        return self._dna._entity(self._parent)

    @property
    def parent_id(self):
        """Returns parent entity identifier"""

        return self._parent

    @parent_id.setter
    def parent_id(self, value):
        """Set parent identifier

        Args:
            value (int): identifier
        Raises:
            ValueError: if parent is not int
        """

        if not isinstance(value, int):
            raise ValueError('Parent id must be int')

        self._parent = value
        self._changed()

        self._dna._update_field(self, 'parent')

    @property
    def name(self):
        """Name of entity"""

        return self._name

    @name.setter
    def name(self, name):
        """Name of entity"""

        self._name = name
        self._changed()

        self._dna._update_field(self, 'name')

    @property
    def type(self):
        """Entity type"""

        return self._type

    @type.setter
    def type(self, value):
        """Set type of entity

        Args:
            value: type of entity
        Raises:
            ValueError: if given value is not int
        """

        if not isinstance(value, int):
            raise ValueError('Type must be int')

        self._type = value
        self._changed()

        self._dna._update_field(self, 'type')

    @property
    def created(self):
        """Time in milliseconds when entity was created"""

        return self._created

    @property
    def modified(self):
        """Time in milliseconds when entity was modified last time"""

        return self._modified

    @property
    def content(self):
        """Get contents of item"""

        return self._content

    @content.setter
    def content(self, content):
        """Set content of item

        Args:
            content (object): contents object
        """

        self._content = content
        self._changed()

        self._dna._update_field(self, 'content')

    @property
    def search(self):
        """Search string"""

        return self._search

    @search.setter
    def search(self, value):
        """Set value of search field

        Args:
            value (str): search string
        Raises:
            ValueError: if given value is not string
        """

        if not isinstance(value, str):
            raise ValueError('Search value must be string')

        self._search = str(value)
        self._changed()

        self._dna._update_field(self, 'search')

    @property
    def index(self):
        """Index inside parent entity"""

        return self._index

    @index.setter
    def index(self, value):
        """Set index of entity

        Args:
            value (int): index value
        """

        if not isinstance(value, int):
            raise ValueError('Index value must be int')

        self._index = value
        self._changed()

        self._dna._update_field(self, 'sort_order')

    def get(self, key, default=None):
        """Get a property value

        Args:
            key (str): property name
            default (object): return `default` if property doesn't exists
        """

        return self._dna._get(self.id, key, default)

    def set(self, key, value, force_type=None):
        """Set a property

        Args:
            key (str): property name
            value (object): value of property
            force_type: convert value to this type
        """

        self._dna._set(self.id, key, value, force_type)

    def has(self, key):
        """Check if property exists

        Args:
            key (str): property name
        """

        return self._dna._has(self.id, key)

    def rename(self, old_key, new_key):
        """Rename property

        Args:
            old_key (str): old property name
            new_key (str): new property name
        """

        return self._dna._rename(self.id, old_key, new_key)

    def unset(self, key):
        """Remove property

        Args:
            key (str): property name
        """

        self._dna._unset(self.id, key)

    def properties(self):
        """Returns list of all properties"""

        return self._dna._properties(self.id)

    def update(self):
        """Force entity update and commit changes to database"""

        self._dna._update(self)

    def delete(self):
        """Delete this entity from DNA"""

        self._dna._remove(self._id)

    def create(self, name, entity_type=None, index=-1, properties=None, factory=None):
        """Create new child entity

        Args:
            name (str): name of entity
            entity_type (int): entity type
            index (int): index where entity will be inserted
            properties (dict): properties list
            factory: factory
        """

        return self._dna._create(name=name,
                                 parent=self._id,
                                 entity_type=entity_type,
                                 index=index,
                                 properties=properties,
                                 factory=factory)

    def remove(self, entity):
        """Remove first entity if given `entity` is children

        Args:
            entity (DNAEntity): entity to be deleted
        """

        self._dna._remove(entity.id, parent=self.id)

    def insert(self, index, entity, factory=None):
        """Create new entity and insert as child item

        Args:
            index (int): index where to insert
            entity (DNAEntity): entity object
            factory: object factory
        """

        return self._dna._copy(entity,
                               parent=self._id,
                               index=index,
                               factory=factory)

    def append(self, entity, factory=None):
        """Create and append new entity

        Args:
            entity (DNAEntity): entity to append
            factory: object factory
        """

        return self._dna._copy(entity, parent=self._id, factory=factory)

    def clear(self):
        """Remove all child entities"""

        self._dna._remove_childs(self._id)

    def childs(self):
        """Returns list of all child entities"""

        return self._dna._childs(self._id)

    def has_childs(self):
        """Returns True if this entity has childs"""

        return self._dna._has_childs(self._id)

    def reverse(self):
        """Sort child entities in reverse order"""

        self._dna._sort(self._id, key='sort_order', reverse=True)

    def sort(self, key=None, reverse=False):
        """Reorder child entities

        Args:
            key (str): database key to sort by
            reverse (bool): reverse sorting
        """

        self._dna._sort(self._id, key=key, reverse=reverse)

    def _changed(self):
        """Called when property is modified"""

        self._modified = millis_now()
        self._dna._update_field(self, 'modified')

    def _parse(self, row):
        """Parse sqlite row into DNAEntity

        Args:
            row (sqlite3.Row): row object
        """

        self._id = int(row[0])
        self._parent = int(row[1])
        self._type = int(row[2])
        self._name = str(row[3])
        self._created = int(row[4])
        self._modified = int(row[5])
        self._content = None
        self._search = str(row[7])
        self._index = int(row[8] or 0)

        if row[6]:
            try:
                self._content = json.loads(str(row[6]))
            except ValueError:
                pass

    @classmethod
    def from_sqlite(cls, parent, row):
        """Parse entity from sqlite"""

        entity = cls(parent=parent)
        entity._parse(row)

        return entity


class SettingsEntity(DNAEntity):
    """Settings object"""

    def __init__(self, parent):
        """Initialize Settings entity

        Args:
            parent (DNA): parent DNA object
        """

        super(SettingsEntity, self).__init__(parent)


class SongEntity(DNAEntity):
    """Representation of song inside Grail file"""

    def __init__(self, parent):
        super(SongEntity, self).__init__(parent)

        # grail songs unique id
        self._uid = 0
        self._name = "Untitled"
        self._artist = "Unknown"
        self._album = "Unknown"
        self._artwork = None
        self._track = 1
        self._year = 0
        self._genre = "Unknown"
        self._language = "en"
        self._lyrics = ""

    @property
    def year(self):
        """Year of song release"""

        return self._year

    @year.setter
    def year(self, value):
        """Set year

        Args:
            value (int): year
        """

        self._year = int(value)
        self._modified = millis_now()

    @property
    def artist(self):
        """Artist name"""

        return self._artist

    @artist.setter
    def artist(self, value):
        """Set artist

        Args:
            value (str): artist name
        """

        self._artist = value
        self._modified = millis_now()

    @property
    def album(self):
        """Album name"""

        return self._album

    @album.setter
    def album(self, value):
        """Set album name

        Args:
            value (str): album name
        """

        self._album = value
        self._modified = millis_now()

    @property
    def track(self):
        """Track number"""

        return self._track

    @track.setter
    def track(self, value):
        """Set track number

        Args:
            value (int): track number
        """

        self._track = int(value)
        self._modified = millis_now()

    @property
    def genre(self):
        """Genre name"""

        return self._genre

    @genre.setter
    def genre(self, value):
        """Set genre

        Args:
            value (str): genre name
        """

        self._genre = value
        self._modified = millis_now()

    @property
    def language(self):
        """Language code of lyrics"""

        return self._language

    @language.setter
    def language(self, value):
        """Set language

        Args:
            value (str): language code
        """

        self._language = value
        self._modified = millis_now()

    @property
    def lyrics(self):
        """Lyrics text"""

        return self._lyrics

    @lyrics.setter
    def lyrics(self, value):
        """Set lyrics text

        Args:
            value (str): new lyrics
        """

        self._lyrics = value
        self._modified = millis_now()

    def update(self):
        """Update song information"""

        self._search = "%s %s %s" % (self._lyrics, self._album, self._artist)
        self._content = {
            'year': self._year,
            'track': self._track,
            'album': self._album,
            'artwork': self._artwork,
            'language': self._language,
            'lyrics': self._lyrics,
            'genre': self._genre,
            'artist': self._artist
            }

        super(SongEntity, self).update()

    def _parse(self, row):
        """Parse sqlite row into SongEntity

        Args:
            row (sqlite3.Row): row object
        """

        super(SongEntity, self)._parse(row)

        content = self._content

        self._year = int(default_key(content, 'year', default=2000))
        self._track = int(default_key(content, 'track', default=1))
        self._album = default_key(content, 'album', 'Unknown')
        self._artwork = None
        self._language = default_key(content, 'language', '')
        self._lyrics = default_key(content, 'lyrics', '')
        self._genre = default_key(content, 'genre', 'Unknown')
        self._artist = default_key(content, 'artist', 'Unknown')


class FileEntity(DNAEntity):
    """Representation of file inside DNA file"""

    SEEK_SET = 0
    SEEK_CUR = 1
    SEEK_END = 2

    def __init__(self, parent):
        super(FileEntity, self).__init__(parent)

        self._offset = 0

    @property
    def content(self):
        """Get file contents"""

        return self._content.encode('utf-8')

    @content.setter
    def content(self, value):
        """Set file contents

        Args:
            value (bytes): new file contents
        """

        if type(value) != bytes:
            raise DNAError('File contents must be bytes object')

        self._content = value.decode('utf-8')

    def read(self, size=0):
        """Read contents of file, if buffer_size is 0 or None returns whole file

        Args:
            size (int): number of bytes to read
        """

        if size == 0 or size is None:
            return self.content

        self._offset += size
        return self.content[self._offset - size:self._offset]

    def write(self, data):
        """Write data at cursor position"""

        content = self.content

        self._content = (content[0:self._offset] + data + content[self._offset + len(data):]).decode('utf-8')
        self._offset += len(data)

    def truncate(self, size=0):
        """Truncates the file's size. If the optional size argument is present,
        the file is truncated to that size.

        Args:
            size (int): size of resulting file
        """

        self._content = self.content[0:size].decode('utf-8')

    def seek(self, offset, whence=SEEK_SET):
        """Sets the file's current position

        Args:
            offset (int): new position
            whence (int): how to calculate offset - from begging, current position or end of contents
        """

        if whence == self.SEEK_SET:
            self._offset = offset
        elif whence == self.SEEK_CUR:
            self._offset += offset
        elif whence == self.SEEK_END:
            self._offset = len(self.content) - offset

    def tell(self):
        """Returns current stream position"""

        return self._offset

    def _parse(self, row):
        """"""

        super(FileEntity, self)._parse(row)

        self._content = (self._content or '')


class CuelistEntity(DNAEntity):
    """Representation of cuelist"""

    def __init__(self, parent):
        """Create a cuelist"""
        super(CuelistEntity, self).__init__(parent)

    def __len__(self):
        """Return number of cues"""

        return len(self.cues())

    def cues(self):
        """Returns list of all cues in Cuelist"""

        return self._dna._entities(filter_parent=self._id)

    def cue(self, cue_id):
        """Get cue by id

        Args:
            cue_id (int): cue identifier
        """

        return self._dna._entity(cue_id)

    def create(self, name, entity_type=None, index=-1, properties=None, factory=None):
        """Create entity with DNA.TYPE_CUE by default

        Args:
            name (str): name of entity
            entity_type (int): entity type
            index (int): index where entity will be inserted
            properties (dict): properties list
            factory: factory
        """

        if entity_type is None:
            entity_type = DNA.TYPE_CUE

        super(CuelistEntity, self).create(name, entity_type, index=-1, properties=properties, factory=factory)


class CueEntity(DNAEntity):
    """Representation of a Cue in Cuelist"""

    # Follow type
    FOLLOW_OFF = 0
    FOLLOW_ON = 1
    FOLLOW_CONTINUE = 2
    FOLLOW_TYPE = (FOLLOW_OFF, FOLLOW_ON, FOLLOW_CONTINUE)

    # default colors
    COLOR_RED = "#B71C1C"
    COLOR_ORANGE = "#EF6C00"
    COLOR_YELLOW = "#FFEB3B"
    COLOR_GREEN = "#8BC34A"
    COLOR_BLUE = "#03A9F4"
    COLOR_PURPLE = "#673AB7"
    COLOR_GRAY = "#BDBDBD"
    COLOR_DEFAULT = "#FFFFFF"

    COLORS = (COLOR_DEFAULT,
              COLOR_RED,
              COLOR_ORANGE,
              COLOR_YELLOW,
              COLOR_GREEN,
              COLOR_BLUE,
              COLOR_PURPLE,
              COLOR_GRAY)

    COLOR_NAMES = (
        'Default',
        'Red',
        'Orange',
        'Yellow',
        'Green',
        'Blue',
        'Purple',
        'Gray'
        )

    def __init__(self, parent):
        """Create a cue instance

        Args:
            parent (DNA): parent DNA
        """
        super(CueEntity, self).__init__(parent)

    @property
    def number(self):
        """Identifier of cue assigned by user"""

        return self.get("number", default=self.index)

    @number.setter
    def number(self, value):
        """Set number

        Args:
            value (str): cue number
        Raises:
            ValueError: if given value is not string
        """

        if not isinstance(value, str):
            raise ValueError('Cue number must be string, \'%s\' given' % str(value))

        self.set("number", value)

    @property
    def color(self):
        """any text that describe cue"""

        return self.get("color", default=self.COLOR_DEFAULT)

    @color.setter
    def color(self, value):
        """Set color of cue

        Args:
            value (str): cue color
        """

        if not isinstance(value, str):
            raise ValueError('Cue color must be string in #000000 format, \'%s\' given' % str(value))

        self.set('color', value)

    @property
    def pre_wait(self):
        """Wait before execution of cue"""

        return self.get("pre-wait", default=0)

    @pre_wait.setter
    def pre_wait(self, value):
        """Set pre wait

        Args:
            value (float, int): pre wait in seconds
        Raises:
            ValueError if given value is not float or int
        """

        if not (isinstance(value, int) or isinstance(value, float)):
            raise ValueError('Given value must be of int or float type')

        self.set('pre-wait', value, force_type=DNA.ARG_FLOAT)

    @property
    def post_wait(self):
        """Wait after execution of cue"""

        return self.get("post-wait", default=0)

    @post_wait.setter
    def post_wait(self, value):
        """Set post wait time

        Args:
            value (float, int): post wait in seconds
        Raises:
            ValueError if given value is not float or int
        """

        if not (isinstance(value, int) or isinstance(value, float)):
            raise ValueError('Given value must be of int or float type')

        self.set('post-wait', value, force_type=DNA.ARG_FLOAT)

    @property
    def follow(self):
        """Execute next cue after this finishes, move cursor to next or do nothing."""

        return self.get("follow", default=self.FOLLOW_OFF)

    @follow.setter
    def follow(self, value):
        """Set follow mode

        Args:
            value (int): follow mode
        Raises:
            ValueError if value is not one of (FOLLOW_OFF, FOLLOW_ON, FOLLOW_CONTINUE)
        """

        if value not in self.FOLLOW_TYPE:
            raise ValueError('Follow type is not supported')

        self.set("follow", value)

    def create(self, name, entity_type=None, index=-1, properties=None, factory=None):
        """Create entity with DNA.TYPE_CUE by default

        Args:
            name (str): name of entity
            entity_type (int): entity type
            index (int): index where entity will be inserted
            properties (dict): properties list
            factory: factory
        """

        if entity_type is None:
            entity_type = DNA.TYPE_CUE

        super(CueEntity, self).create(name, entity_type, index=-1, properties=properties, factory=factory)


class DNA:
    """Base class for parsing grail files,
    use this class when you don't want to expose internal methods.
    Otherwise use DNAFile class.
    Please don't use this class directly.
    """

    # property types
    ARG_NONE = 0
    ARG_BOOL = 1
    ARG_INT = 2
    ARG_FLOAT = 3
    ARG_STRING = 4
    ARG_JSON = 6

    # list all supported types
    _SUPPORTED_ARGS = (ARG_NONE, ARG_BOOL, ARG_INT, ARG_FLOAT, ARG_STRING, ARG_JSON)

    # types of entities
    TYPE_ABSTRACT = 0
    TYPE_BIBLE = 10
    TYPE_VERSE = 11
    TYPE_BOOK = 12
    TYPE_PROJECT = 20
    TYPE_SETTINGS = 21
    TYPE_CUELIST = 22
    TYPE_CUE = 23
    TYPE_LIBRARY = 24
    TYPE_FILE = 25
    TYPE_SONG = 26
    TYPE_LAYOUT = 31
    TYPE_VIEW = 32

    # enumerate all types of entities
    TYPES = (
        TYPE_ABSTRACT,
        TYPE_BIBLE,
        TYPE_VERSE,
        TYPE_BOOK,
        TYPE_PROJECT,
        TYPE_SETTINGS,
        TYPE_CUELIST,
        TYPE_CUE,
        TYPE_LIBRARY,
        TYPE_FILE,
        TYPE_SONG,
        TYPE_LAYOUT,
        TYPE_VIEW
        )

    # types and their default factories
    TYPES_FACTORIES = {
        TYPE_ABSTRACT: DNAEntity,
        TYPE_SETTINGS: SettingsEntity,
        TYPE_CUELIST: CuelistEntity,
        TYPE_CUE: CueEntity,
        TYPE_FILE: FileEntity,
        TYPE_SONG: SongEntity
        }

    # database handler
    _db = None
    # create file from this query
    _db_create_query = """
        DROP TABLE IF EXISTS properties;
        CREATE TABLE properties(entity INT, key TEXT, value TEXT, type INT);

        DROP TABLE IF EXISTS entities;
        CREATE TABLE entities(id INTEGER PRIMARY KEY AUTOINCREMENT, parent INT, type INT, name TEXT,
                              created INT, modified INT, content BLOB, search TEXT, sort_order INT);
        """
    # file extension
    _file_extension = ".grail"

    def __init__(self, file_path, create=False):
        """Open a *.grail file for read and write

        Args:
            file_path (str): path to file
            create (bool): create file if not exists

        Raises:
            DNAError: if file can't be parsed or not exists
        """

        # signals
        self.property_changed = Signal(int, str, str)
        self.entity_changed = Signal(int)
        self.entity_removed = Signal(int)
        self.entity_added = Signal(int)

        self._changed = False
        self._location = file_path

        if not self.validate(file_path) and not create:
            raise DNAError("Grail file '%s' could not be opened." % file_path)
        else:
            self._db = DataBaseHost.get(file_path, query=self._db_create_query, create=create)

    @property
    def location(self):
        """Returns path to file"""

        return self._location

    @property
    def filename(self):
        """Returns file name"""

        return os.path.splitext(os.path.basename(self._location))[0]

    @property
    def changed(self):
        """Return True if changes not saved"""

        return self._changed

    def save(self):
        """Save all changes"""

        self._db.commit()
        self._changed = False

    def save_copy(self, file_path, create=True):
        """Save a copy of this file to `file_path` location

        Args:
            file_path (str): path to copy of current file
            create (bool): If True file will be created
        """

        self.save()
        self._db.copy(file_path, create=create)

    def close(self):
        """Close connection"""

        self._db.close()
        self._changed = False

    def _create(self, name="", parent=0, entity_type=None, index=-1, properties=None, factory=None):
        """Returns new DNAEntity inside this DNA file

        Args:
            name (str): entity name
            parent (int): parent id
            entity_type (int, None): entity type
            index (int): index of insertion
            properties (dict, None): dict with entity properties
            factory: object creation factory
        Returns:
            DNAEntity by default or if `factory` given returns object created by factory
        """

        self._changed = True

        if entity_type not in DNA.TYPES:
            entity_type = DNA.TYPE_ABSTRACT

        cursor = self._db.cursor
        sql_values = [parent,
                      entity_type,
                      name,
                      millis_now(),
                      millis_now(),
                      json.dumps(None, separators=(',', ':')),
                      '',
                      index]

        if index >= 0:
            # realign sort_order field
            cursor.execute("""UPDATE entities
                              SET sort_order = sort_order + 1
                              WHERE sort_order >= ? AND parent = ?""", (index, parent))
            # insert
            cursor.execute("""INSERT INTO entities
                          (id, parent, type, name, created, modified, content, search, sort_order)
                          VALUES (NULL, ?, ?, ?, ?, ?, ?, ?, ?)""", sql_values)
        else:
            sql_values[7] = parent
            # append item with max(sort_order) + 1 index or 0 if it first child entity
            cursor.execute("""INSERT INTO entities
                          (id, parent, type, name, created, modified, content, search, sort_order)
                          VALUES (NULL, ?, ?, ?, ?, ?, ?, ?,
                          (SELECT IFNULL(MAX(sort_order) + 1, 0) FROM entities WHERE parent = ?))
                           """, sql_values)

        self._db.connection.commit()

        entity_id = cursor.lastrowid

        if properties and len(properties) > 0:
            for key, value in properties.items():
                force_type = self._get_type(value)
                self._db.execute("INSERT OR IGNORE INTO properties(entity, key, value, type) VALUES(?, ?, ?, ?)",
                                 (entity_id, key, self._write_type(value, force_type), force_type))

        entity = self._entity(entity_id, factory)

        # emit entity changed
        self.entity_changed.emit(entity.id)
        self.entity_added.emit(entity.id)

        return entity

    def _copy(self, entity, name=None, parent=None, entity_type=None, index=-1, factory=None):
        """Create copy of given entity and override properties if needed

        Args:
            entity (DNAEntity): entity to be copied
            name (str): override name of entity
            parent (int): override entity parent
            entity_type (int): override entity type
            index (int): override entity index
            factory: create entity using this factory
        """

        name = name if name else entity.name
        parent = parent if parent else entity.parent_id
        entity_type = entity_type if entity_type else entity.type

        cursor = self._db.cursor
        sql_values = [parent,
                      entity_type,
                      name,
                      millis_now(),
                      millis_now(),
                      json.dumps(entity.content, separators=(',', ':')),
                      entity.search,
                      index]

        if index >= 0:
            # realign sort_order field
            cursor.execute("""UPDATE entities
                              SET sort_order = sort_order + 1
                              WHERE sort_order >= ? AND parent = ?""", (index, parent))
            # insert
            cursor.execute("""INSERT INTO entities
                          (id, parent, type, name, created, modified, content, search, sort_order)
                          VALUES (NULL, ?, ?, ?, ?, ?, ?, ?, ?)""", sql_values)
        else:
            sql_values[7] = parent
            # append item with max(sort_order) + 1 index or 0 if it first child entity
            cursor.execute("""INSERT INTO entities
                          (id, parent, type, name, created, modified, content, search, sort_order)
                          VALUES (NULL, ?, ?, ?, ?, ?, ?, ?,
                          (SELECT IFNULL(MAX(sort_order) + 1, 0) FROM entities WHERE parent = ?))
                           """, sql_values)

        self._db.connection.commit()

        copy_id = cursor.lastrowid

        for key, value in entity.properties().items():
            force_type = self._get_type(value)
            self._db.execute("INSERT OR IGNORE INTO properties(entity, key, value, type) VALUES(?, ?, ?, ?)",
                             (copy_id, key, self._write_type(value, force_type), force_type))

        entity = self._entity(copy_id, factory=factory)
        self.entity_added.emit(entity.id)

        return entity

    def _entity(self, entity_id, factory=None):
        """Get entity by `entity_id`

        Args:
            entity_id (int): id of an entity
            factory: use another class to create entity

        Returns:
            DNAEntity with id `entity_id` if entity exists otherwise returns Null
        """
        raw_entities = self._db.all("""SELECT id, parent, type, name, created, modified, content, search, sort_order
                            FROM entities WHERE id = ?""", (entity_id,))
        entities = self._factory(factory, raw_entities)

        return entities[0] if len(entities) > 0 else None

    def _entities(self, filter_type=None, filter_parent=None, filter_keyword=None,
                  sort=None, reverse=False, offset=0, limit=0, factory=None):
        """Get list of all entities

        Args:
            filter_type (int): filter by type of entity
            filter_parent (int): filter by id of parent entity
            filter_keyword (str): filter by name
            sort (str): property to sort by
            reverse (bool): reverse order of items
            offset (int): offset a result set
            limit (int): limit result set
            factory (object): factory object

        Returns:
            list of all entities available regarding filter arguments
        """

        where = []
        args = []

        if filter_parent is not None:
            where.append(" parent = ?")
            args.append(filter_parent)

        if filter_type is not None:
            where.append(" type = ?")
            args.append(filter_type)

        # empty keyword may result in selecting all records
        if filter_keyword and len(filter_keyword) > 0:
            keyword = "%" + str(filter_keyword) + "%"
            where.append(""" lowercase(name) LIKE lowercase(?)
                    OR lowercase(search) LIKE lowercase(?)""")
            args.append(keyword)
            args.append(keyword)

        sql = """
            SELECT id, parent, type, name, created, modified, content, search, sort_order
            FROM entities
            %s
            ORDER BY %s %s
            """ % ('WHERE' + ' AND '.join(where) if len(where) > 0 else '',
                   sort if isinstance(sort, str) else "sort_order",
                   "DESC" if reverse else "ASC")

        if limit > 0:
            sql += " LIMIT ? "
            args.append(limit)

        if offset > 0:
            sql += " OFFSET ? "
            args.append(offset)

        raw_entities = self._db.all(sql, args)

        return self._factory(factory, raw_entities)

    def _update(self, entity):
        """update entity

        Args:
            entity (DNAEntity): entity object
        """

        self._changed = True

        cursor = self._db.cursor
        cursor.execute("""UPDATE entities SET
                        id = ?, parent = ?, type = ?, name = ?, created = ?, modified = ?, content = ?, search = ?,
                        sort_order = ?
                        WHERE id = ?""",
                       (entity.id, entity.parent_id, entity.type, entity.name, entity.created, entity.modified,
                        json.dumps(entity._content, separators=(',', ':')), entity.search, entity.index, entity.id))

        self.entity_changed.emit(entity.id)

    def _update_field(self, entity, field):
        """Update entity field

        Args:
            entity (DNAEntity): entity to be changed
            field (str): database field to be updated
        """

        self._changed = True

        if field not in ('parent', 'type', 'name', 'created', 'modified', 'content', 'search', 'sort_order'):
            raise ValueError('Given field "%s" not supported' % str(field))

        if field in ('type', 'name', 'created', 'modified', 'search'):
            value = getattr(entity, field)
        elif field == 'parent':
            value = entity.parent_id
        elif field == 'content':
            value = json.dumps(entity.content, separators=(',', ':'))
        elif field == 'sort_order':
            value = entity.index
        else:
            value = ''

        cursor = self._db.cursor
        cursor.execute("""UPDATE entities SET %s = ? WHERE id = ?""" % field, (value, entity.id))

        self.entity_changed.emit(entity.id)

    def _remove(self, entity_id, parent=None):
        """Remove entity by id

        Args:
            entity_id (int): id of an entity
            parent (int): parent entity id
        """

        if parent:
            entity = self._db.get("SELECT parent FROM entities WHERE id = ? AND parent = ?", (entity_id, parent))
            parent_id = entity[0] if entity else None

            # return False since there is no entity to remove
            if not entity:
                return False
        else:
            parent_id = self._db.get("SELECT parent FROM entities WHERE id = ?", (entity_id,))
            parent_id = parent_id[0] if parent_id else None

        self._changed = True

        self._db.execute("DELETE FROM entities WHERE id = ?", (entity_id,))
        self._db.execute("DELETE FROM properties WHERE entity = ?", (entity_id,))
        self._db.connection.commit()

        # emit entity changed signal
        # emit parent id, as we can't use removed entity
        self.entity_removed.emit(parent_id)

        self._remove_childs(entity_id)

        return True

    def _remove_childs(self, entity_id):
        """Remove child entities

        Args:
            entity_id (int): id of DNAEntity
        """

        for child in self._childs(entity_id):
            self._remove(child.id)

    def _sort(self, entity_id, key=None, reverse=False):
        """Sort child entities

        Args:
            entity_id (int): entity id
            key: sort field
            reverse (bool): reverse order or not
        """

        entities = self._entities(filter_parent=entity_id, sort=key, reverse=reverse)

        for index, entity in enumerate(entities):
            self._db.execute("""UPDATE entities SET sort_order = ? WHERE id = ?""", (index, entity.id))

        self._db.commit()

    def _has_childs(self, entity_id):
        """Check entity child nodes

        Args:
            entity_id (int): id of an entity

        Returns:
            True if entity has child nodes
        """
        parent = self._db.get("SELECT id FROM entities WHERE parent = ?", (entity_id,))

        return bool(parent)

    def _childs(self, entity_id, factory=None):
        """Get child nodes of entity

        Args:
            entity_id (id): id of an entity

        Returns:
            list of child entities
        """
        raw_entities = self._db.all("""SELECT id, parent, type, name, created, modified, content, search, sort_order
                                FROM entities
                                WHERE parent = ?
                                ORDER BY sort_order ASC""", (entity_id,))

        return self._factory(factory, raw_entities)

    def _get(self, entity_id, key, default=None):
        """Get property of entity with id `entity_id` and property name `key`

        Args:
            entity_id (int): id of an entity
            key (str): name of property
            default (object): value returned if property is not existed

        Returns:
            Value of property of an entity
        """

        value = self._db.get("SELECT value, type FROM `properties` WHERE `entity` = ? AND `key` = ?", (entity_id, key))

        return self._read_type(value[0], value[1]) if value else default

    def _set(self, entity_id, key, value, force_type=None):
        """Set a property value of an entity

        Args:
            entity_id (int): id of an entity
            key (str): name of property
            value (object): value to assign to property
            force_type (DNA.TYPE): set a type of property
        Returns:
            id of last inserted row
        """
        self._changed = True

        if force_type is None:
            force_type = self._get_type(value)
        elif force_type not in self._SUPPORTED_ARGS:
            raise DNAError("Property type not supported")

        value = self._write_type(value, force_type)

        self._db.execute("INSERT OR IGNORE INTO properties(entity, key, value, type) VALUES(?, ?, ?, ?)",
                         (entity_id, key, value, force_type))
        self._db.execute("UPDATE properties SET entity = ?, key = ?, value = ?, type = ? WHERE entity = ? AND key = ?",
                         (entity_id, key, value, force_type, entity_id, key))

        self.property_changed.emit(entity_id, key, value)

        return self._db.cursor.lastrowid

    def _has(self, entity_id, key):
        """Check property existence

        Args:
            entity_id (int): id of an entity
            key (str): name of property

        Returns:
            True if property exists
        """

        value = self._db.get("SELECT value FROM `properties` WHERE `entity` = ? AND `key` = ?",
                             (entity_id, key))

        return bool(value)

    def _contains(self, entity_id, child_id):
        """Check if entity contains a child

        Args:
            entity_id (int): entity id
            child_id (int): child entity id
        """

        value = False
        raw_childs = self._db.all("SELECT id FROM entities WHERE parent = ?", (entity_id,))

        for child in raw_childs:
            if child[0] == child_id:
                return True

            value = self._contains(child[0], child_id)

        return value

    def _properties(self, entity_id):
        """Get list of all properties linked to entity

        Args:
            entity_id (int): id of an entity

        Returns:
            properties list of an entity
        """

        props = self._db.all("SELECT key, value, type FROM `properties` WHERE `entity` = ?", (entity_id,))

        return {prop[0]: self._read_type(prop[1], prop[2]) for prop in props}

    def _unset(self, entity_id, key):
        """Remove property of an entity

        Args:
            entity_id (int): id of an entity
            key (str): name of property
        """
        self._changed = True

        self._db.execute("DELETE FROM properties WHERE entity = ? AND key = ?", (entity_id, key))

    def _unset_all(self, entity_id):
        """Remove all properties of an entity

        Args:
            entity_id (int): id of an entity
        """
        self._changed = True

        self._db.execute("DELETE FROM properties WHERE entity = ?", (entity_id,))

    def _rename(self, entity_id, old_key, new_key):
        """Rename property key

        Args:
            entity_id (int): id of an entity
            old_key (str): old property key
            new_key (str): new property key
        """

        self._changed = True

        self._db.execute("UPDATE properties SET key = ? WHERE entity = ? AND key = ?",
                         (new_key, entity_id, old_key))

        # emit property changed signal
        self.property_changed.emit(entity_id, new_key, self._get(entity_id, new_key))

    def _factory(self, factory, raw_entities):
        """Return list of entities created by factory from raw_entities

        Args:
            factory: object from which entities will be created
            raw_entities: sqlite3 rows
        """

        entities = []

        for raw in raw_entities:
            factory_entity = factory if factory else self._get_factory(raw['type'])
            entities.append(factory_entity.from_sqlite(self, raw))

        return entities

    @classmethod
    def validate(cls, file_path):
        """Validate file to be proper grail file

        Args:
            file_path (str): path to file
        """

        # check file existence
        if not os.path.isfile(file_path):
            return False

        # check file extension
        if os.path.splitext(file_path)[1] != cls._file_extension:
            return False

        # try to open sqlite database
        try:
            db = DataBaseHost.get(file_path)
            db.all("SELECT name FROM sqlite_master WHERE type='table'")
        except lite.DatabaseError:
            return False

        return True

    @classmethod
    def _get_factory(cls, entity_type):
        """Map types to entities

        Args:
            entity_type (int): type of entity defined by type field
        Returns:
            DNAEntity or DNAEntity subclass corresponding to type as defined in DNA.TYPES_FACTORIES
        """

        return cls.TYPES_FACTORIES[entity_type] if entity_type in cls.TYPES_FACTORIES else DNAEntity

    @classmethod
    def _get_type(cls, value):
        """Get type of value

        Args:
            value: any object
        Returns:
            argument type
        """

        builtin_type = type(value)
        arg_type = cls.ARG_NONE

        if builtin_type == str:
            arg_type = cls.ARG_STRING
        elif builtin_type == int:
            arg_type = cls.ARG_INT
        elif builtin_type == float:
            arg_type = cls.ARG_FLOAT
        elif builtin_type == bool:
            arg_type = cls.ARG_BOOL
        elif builtin_type == dict or builtin_type == list:
            arg_type = cls.ARG_JSON

        return arg_type

    @classmethod
    def _write_type(cls, arg_value, arg_type):
        """Create a string representation of value given

        Args:
            arg_value: any object
            arg_type: type of object to convert to

        Returns:
            string representation of value that will be written to database
        """

        if arg_type == cls.ARG_JSON:
            value = json.dumps(arg_value)
        elif arg_value is True:
            value = 'True'
        elif arg_value is False:
            value = 'False'
        elif arg_type in cls._SUPPORTED_ARGS and arg_type != cls.ARG_JSON:
            value = str(arg_value)
        else:
            value = 'None'

        return value

    @classmethod
    def _read_type(cls, arg_value, arg_type):
        """Parse string from db into correct object

        Args:
            arg_value: raw value
            arg_type: value type

        Returns:
            object from raw value
        """

        if arg_type == cls.ARG_JSON:
            arg_value = json.loads(arg_value)
        elif arg_type == cls.ARG_INT:
            arg_value = int(arg_value)
        elif arg_type == cls.ARG_FLOAT:
            arg_value = float(arg_value)
        elif arg_type == cls.ARG_BOOL:
            arg_value = arg_value == 'True'
        elif arg_type == cls.ARG_NONE:
            arg_value = None

        return arg_value


class DNAProxy:
    """This class gives public access to hidden methods of DNA"""

    def __init__(self, dna):
        """Args:
            dna (DNA): reference to DNA
        """

        self._dna_ref = dna

    def create(self, *args, **kwargs):
        """Create entity"""
        return self._dna_ref._create(*args, **kwargs)

    def copy(self, *args, **kwargs):
        """Copy existing entity"""
        return self._dna_ref._copy(*args, **kwargs)

    def entity(self, *args, **kwargs):
        """Returns entity by id"""
        return self._dna_ref._entity(*args, **kwargs)

    def entities(self, *args, **kwargs):
        """Returns list of entities"""
        return self._dna_ref._entities(*args, **kwargs)

    def update(self, *args, **kwargs):
        """Update entity"""
        return self._dna_ref._update(*args, **kwargs)

    def remove(self, *args, **kwargs):
        """Remove entity by id"""
        return self._dna_ref._remove(*args, **kwargs)

    def childs(self, *args, **kwargs):
        """Returns list of child entities"""
        return self._dna_ref._childs(*args, **kwargs)

    def has_childs(self, *args, **kwargs):
        """Returns Tru if entity has childs"""
        return self._dna_ref._has_childs(*args, **kwargs)

    def contains(self, *args, **kwargs):
        """Check if entity contains an child with given id"""
        return self._dna_ref._contains(*args, **kwargs)

    def has(self, *args, **kwargs):
        """Check if property exists"""
        return self._dna_ref._has(*args, **kwargs)

    def get(self, *args, **kwargs):
        """Get property value"""
        return self._dna_ref._get(*args, **kwargs)

    def set(self, *args, **kwargs):
        """Set property value"""
        return self._dna_ref._set(*args, **kwargs)

    def rename(self, *args, **kwargs):
        """Rename property name"""
        return self._dna_ref._rename(*args, **kwargs)

    def unset(self, *args, **kwargs):
        """Remove property"""
        return self._dna_ref._unset(*args, **kwargs)

    def unset_all(self, *args, **kwargs):
        """Remove all properties"""
        return self._dna_ref._unset_all(*args, **kwargs)

    def properties(self, *args, **kwargs):
        """Returns dict of properties"""
        return self._dna_ref._properties(*args, **kwargs)


class DNAFile(DNA, DNAProxy):
    """Interface to Grail-file with public methods"""

    def __init__(self, file_path, create=False):
        """Open a grail file

        Args:
            file_path (str): path to grail file
            create (bool): create file if not exists
        """
        DNA.__init__(self, file_path, create=create)
        DNAProxy.__init__(self, self)


class SettingsFile(DNA):
    """Represents a grail file with properties only"""

    def __init__(self, file_path, create=False):
        """Open or create a settings file

        Args:
            file_path (str): path to file
            create (bool): create file if not exists
        """

        super(SettingsFile, self).__init__(file_path, create=create)

    def has(self, key):
        """Check if property exists

        Args:
            key (str): property key
        Returns:
            True if property exists
        """

        return self._has(0, key)

    def get(self, key, default=None):
        """Get a property value

        Args:
            key (str): property key
            default (object): Object that will be returned if property doesn't exists
        Returns:
            property value or default if isn't exists
        """

        return self._get(0, key, default)

    def set(self, key, value, force_type=None):
        """Set value of property

        Args:
            key (str): property key
            value (object): property value object
            force_type: If set to not None, value with this type will be saved
        Returns:
            True if operation is succeeded
        """

        result = self._set(0, key, value, force_type)
        self._db.commit()

        return result

    def properties(self):
        """Get a all properties

        Returns: dist object of all properties
        """

        return self._properties(0)

    def unset(self, key):
        """Remove property by key value

        Args:
            key (str): property key
        Returns:
            True if operation is succeeded
        """

        self._unset(0, key)
        self._db.commit()

        return True

    def unset_all(self):
        """Remove all properties

        Returns:
            True if operation succeeded
        """

        self._unset_all(0)
        self._db.commit()

        return True

    def rename(self, old_key, new_key):
        """Rename property key

        Args:
            old_key (str): current property key
            new_key (str): new property key

        Returns:
            True if operation is succeeded
        """

        self._rename(0, old_key, new_key)
        self._db.commit()

        return True


class ProjectError(DNAError):
    """Base class for all project related exceptions"""

    pass


class Project(DNA):
    """Representation of a project file"""

    def __init__(self, file_path, create=False):
        """Open or create a project

        Args:
            file_path (str): path to file
            create (bool): create file if not exists
        """

        super(Project, self).__init__(file_path, create=create)

        self._dna_proxy = DNAProxy(self)
        self._project = 0

        root = self._root()

        if not create and not root:
            raise ProjectError('Unable to open project. Root entity not found.')

        # create project if there is no project entity
        if create and not root:
            self._create_project()

        self._id = self._root().id

    def __len__(self):
        """Get number of cuelist in project"""

        return len(self.cuelists())

    @property
    def dna(self):
        """Get a proxy for dna"""

        return self._dna_proxy

    @property
    def name(self):
        """Project name"""

        return self._root().name

    @name.setter
    def name(self, value):
        """Project name setter

        Args:
            value (str): new project name
        """

        self._root().name = value

    @property
    def description(self):
        """Project description"""

        return self._root().get('description', default='Grail project')

    @description.setter
    def description(self, value):
        """Project description setter

        Args:
            value (str): new project description
        """

        self._root().set('description', value)

    @property
    def author(self):
        """Project author"""

        return self._root().get('author', default='Grail')

    @author.setter
    def author(self, value):
        """Project author setter

        Args:
            value (str): new project author
        """

        self._root().set('author', value)

    @property
    def created(self):
        """Date on witch project was created"""

        return self._root().created

    @property
    def modified(self):
        """Date of last edit on project"""

        return self._root().modified

    def settings(self):
        """Get a setting entity"""

        entity = self._entities(filter_type=DNA.TYPE_SETTINGS,
                                filter_parent=self._id,
                                factory=SettingsEntity)

        if len(entity) == 0:
            entity = self._create(name="Settings",
                                  parent=self._id,
                                  entity_type=DNA.TYPE_SETTINGS,
                                  factory=SettingsEntity)
            return entity
        else:
            return entity[0]

    def cuelists(self):
        """Get all cuelists in project"""

        return self._entities(filter_type=DNA.TYPE_CUELIST,
                              filter_parent=self._id,
                              factory=CuelistEntity)

    def cuelist(self, cuelist_id):
        """Get a cuelist"""

        return self._entity(cuelist_id, factory=CuelistEntity)

    def remove(self, cuelist_id):
        """Remove a cuelist"""

        return self._remove(cuelist_id)

    def create(self, name="Untitled cuelist"):
        """Create a cuelist"""

        return self._create(name=name, parent=self._id, entity_type=DNA.TYPE_CUELIST, factory=CuelistEntity)

    def _root(self):
        """Returns root project entity"""

        if self._project:
            return self._project

        root = self._entities(filter_type=DNA.TYPE_PROJECT, filter_parent=0)
        root = root[0] if len(root) > 0 else None
        self._project = root

        return root

    def _create_project(self):
        """Create project entities"""

        self._project = self._create(name='Untitled project',
                                     parent=0,
                                     entity_type=DNA.TYPE_PROJECT,
                                     properties={
                                      'author': 'Grail',
                                      'description': 'Grail project'
                                     })
        self._id = self._project._id
        self._create(name="settings", entity_type=DNA.TYPE_SETTINGS, parent=self._id)


class LibraryError(DNAError):
    """Library error of any kind"""

    pass


class Library(DNA):
    """Manage library file"""

    # file extension
    _file_extension = ".grail-library"

    def __init__(self, file_path, create=False):
        """Open or create a project

        Args:
            file_path (str): path to file
        """
        super(Library, self).__init__(file_path, create=create)

        self._root = None
        self._dna_proxy = DNAProxy(self)

        if not self.root() and create:
            self._create("Grail Library", parent=0, entity_type=DNA.TYPE_LIBRARY)

        if self.root() is None:
            raise LibraryError("Library entity not found in file %s" % (file_path,))

    @property
    def dna(self):
        """Get a proxy for dna"""

        return self._dna_proxy

    def create(self, name, entity_type=DNA.TYPE_ABSTRACT, factory=None):
        """Create new library entity

        Returns: new item
        """

        item = self._create(name,
                            parent=self.root().id,
                            entity_type=entity_type,
                            factory=factory)

        return item

    def copy(self, entity):
        """Copy existing entity into library

        Args:
            entity (DNAEntity): entity to be copied
        """

        self._copy(entity)

    def root(self):
        """Get a library entity

        Returns: root item of library
        """

        if self._root:
            return self._root

        root = self._entities(filter_type=DNA.TYPE_LIBRARY, filter_parent=0)
        root = root[0] if len(root) > 0 else None
        self._root = root

        return root

    def remove(self, entity_id):
        """Remove entity from library

        Args:
            entity_id (int): id of entity
        """

        self._remove(entity_id)

    def clear(self):
        """Remove all entities from library"""

        for entity in self.items():
            self._remove(entity.id)

    def items(self, filter_type=None, filter_keyword=None, sort='name', reverse=False, offset=0, limit=0):
        """Returns list of library items

        Args:
            filter_type: limit result set by type, pass False to disable filter
            filter_keyword (str): limit result set by keyword, pass False to disable filter
            sort (str): sort field
            reverse (bool): reverse results order
            offset (int): start index
            limit (int): limit items result set
        """

        return self._entities(
            filter_type=filter_type,
            filter_parent=self.root().id,
            filter_keyword=filter_keyword,
            sort=sort,
            reverse=reverse,
            offset=offset,
            limit=limit)

    def item(self, entity_id):
        """Return library item by id

        Args:
            entity_id (int): entity identifier
        """

        return self._entity(entity_id)



# -*- coding: UTF-8 -*-
"""
    grailkit.plug
    ~~~~~~~~~~~~~

    Simple plugin discovery, loading & registration mechanisms

    :copyright: (c) 2017-2019 by Oleksii Lytvyn.
    :license: MIT, see LICENSE for more details.
"""
import re
import os
import importlib


class PluginRegistry(type):
    """Subclasses registration mechanism based on meta-classes

    Use this class as meta-class to register plugin classes
    """

    def __init__(cls, *args):
        """Args:
            name (str): class name
            bases (list): class bases
            attrs (list): attributes
        """
        super(PluginRegistry, cls).__init__(*args)

        if not hasattr(cls, '__registry__'):
            cls.__registry__ = set()

        cls.__registry__.add(cls)

        # Remove base classes
        cls.__registry__ -= set(args[1])


class Plugin(object, metaclass=PluginRegistry):
    """Base plugin class"""

    def __init__(self):
        pass

    @classmethod
    def plugins(cls):
        """Returns list of classes extended from Plugin"""

        return cls.__registry__


def discover(location, packages=False, exclude=None):
    """Load and execute python modules in given location

    Args:
        location (str): path to plugins folder
        packages (bool): If True sub-modules will be also loaded
        exclude (list): list of excluded modules, if modules also included. see `include_modules` argument
    Returns:
        list: modules loaded
    """

    location = os.path.abspath(location)
    module_name = os.path.basename(location)
    files = filter(re.compile('.py$', re.IGNORECASE).search, os.listdir(location))
    plugins = map(lambda fp: '.' + os.path.splitext(fp)[0], files)

    # import parent module / namespace
    importlib.import_module(module_name)
    modules = []
    exclude = exclude or []

    # include packages
    if packages:
        for name in os.listdir(location):
            if os.path.isfile(os.path.join(location, name + '/__init__.py')) and name not in exclude:
                modules.append(importlib.import_module(name, package=module_name))

    # include modules
    for plugin in plugins:
        if not plugin.startswith('__'):
            modules.append(importlib.import_module(plugin, package=module_name))

    return modules



# -*- coding: UTF-8 -*-
"""
    grailkit
    ~~~~~~~~

    GrailKit is a library for Grail like apps,
    library which consists of UI components for PyQt,
    grail files IO and networking libraries

    :copyright: (c) 2017-2019 by Oleksii Lytvyn.
    :license: MIT, see LICENSE for more details.
"""

import os
from grailkit.util import data_location

# library version
__version__ = '0.8.1'

# path to shader folder
PATH_SHARED = data_location("grail-shared")

# path to library
PATH_LIBRARY = os.path.join(PATH_SHARED, "library.grail-library")

# path to settings
PATH_SETTINGS = os.path.join(PATH_SHARED, "settings.grail")



# -*- coding: UTF-8 -*-
"""
    setup
    ~~~~~

    :copyright: (c) 2017-2019 by Oleksii Lytvyn.
    :license: MIT, see LICENSE for more details.
"""

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

import grailkit


setup(
    name='grailkit',
    version=grailkit.__version__,
    author='Oleksii Lytvyn',
    author_email='programer95@gmail.com',
    description='Grail development kit, a set of libraries for creative coding',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url='https://bitbucket.org/alexlitvin/grailkit',
    download_url='https://bitbucket.org/alexlitvin/grailkit/get/default.zip',
    platforms='any',
    packages=['grailkit'],
    keywords=['framework', 'grail', 'development', 'osc', 'midi', 'dmx', 'utilities', 'spout'],
    zip_safe=False,
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Intended Audience :: Information Technology',
        'Intended Audience :: Telecommunications Industry',
        'Operating System :: OS Independent',
        'Environment :: Other Environment',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.0',
        'Programming Language :: Python :: 3.1',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: Implementation :: CPython',
        'Topic :: System :: Networking',
        'Topic :: Multimedia :: Sound/Audio',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'License :: OSI Approved :: MIT License'
    ],
    install_requires=['pyserial', 'python-rtmidi']
)
